'''
CourtID: ill
Court Short Name: Ill.
Author: Rebecca Fordon
Reviewer: Mike Lissner
History:
* 2016-06-22: Created by Rebecca Fordon
'''


from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OralArgumentSite import OralArgumentSite


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.illinoiscourts.gov/Media/On_Demand.asp'
        self.download_url_path = "/td[6]//@href"
        self.case_name_path = '/td[3]'
        self.docket_number_path = "/td[2]"
        # Extract data from all rows with mp3 url/link
        self.xpath_root = '(//table[(.//th)[1][contains(.//text(), "Argument Date")]])[last()]//tr[position() > 1][.//@href[contains(., ".mp3")]]'
        self.back_scrape_iterable = range(2008, 2016)

    def _get_download_urls(self):
        path = self.xpath_root + self.download_url_path
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        dates = []
        path = self.xpath_root + "/td[1]"
        for e in self.html.xpath(path):
            s = e.text_content()
            try:
                d = convert_date_string(s, fuzzy=True)
            except ValueError:
                continue
            else:
                dates.append(d)
        return dates

    def _get_case_names(self):
        path = self.xpath_root + self.case_name_path
        cases = []
        for case in self.html.xpath(path):
            case = case.text_content().strip()
            if case:
                cases.append(case)
        return cases

    def _get_docket_numbers(self):
        path = self.xpath_root + self.docket_number_path
        dockets = []
        for docket in self.html.xpath(path):
            docket = docket.text_content()
            dockets.append(docket)
        return dockets

    def _download_backwards(self, date_str):
        """Backwards urls are just the regular ones with a year munged at the
        end.
        """
        # Set it to the value that seems to work everywhere.
        if getattr(self, 'orig_url', None):
            # Set url back to its original value, if it has been reset already.
            self.url = self.orig_url
        else:
            # First iteration. Set aside the original value, so we can use it
            # in the remaining iterations.
            self.orig_url = self.url

        parts = self.url.rsplit('.', 1)
        self.url = "%s_%s.%s" % (parts[0], date_str, parts[1])
        self.html = self._download()



# Author: Michael Lissner
# Date created: 2013-06-13

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://supremecourt.nebraska.gov/courts/supreme-court/opinions'

    def _get_download_urls(self):
        return [href for href in self.html.xpath('//td[3]/a/@href')]

    def _get_case_names(self):
        return [cell.text_content() for cell in self.html.xpath('//td[3]')]

    def _get_case_dates(self):
        dates = []
        path_group = '//div[@class="view-grouping"]'
        subpath_date = './/span[@class="date-display-single"]'
        subpath_cases = './/tbody/tr'
        for group in self.html.xpath(path_group):
            date_string = group.xpath(subpath_date)[0].text_content()
            date = convert_date_string(date_string)
            case_count = len(group.xpath(subpath_cases))
            dates.extend([date] * case_count)
        return dates

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        return [cell.text_content() for cell in self.html.xpath('//td[1]')]

    def _get_neutral_citations(self):
        return [cell.text_content() for cell in self.html.xpath('//td[2]')]


"""
History:
    - 2014-08-05: Updated by mlr because it was not working, however, in middle
      of update, site appeared to change. At first there were about five
      columns in the table and scraper was failing. Soon, there were seven and
      the scraper started working without my fixing it. Very odd.
"""

from datetime import date
from dateutil.rrule import DAILY, rrule

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase, convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.ca9.uscourts.gov/opinions/index.php"
        self.position = '[position() > 1]'
        self.base = self.get_base_path()
        self.back_scrape_date_start = date(2005, 1, 3)
        self.back_scrape_date_end = date(2015, 1, 1)
        self.back_scrape_iterable = self.get_backscrape_iterable()
        self.back_scrape_url = self.url

    def get_backscrape_iterable(self):
        date_rules = rrule(DAILY, dtstart=self.back_scrape_date_start, until=self.back_scrape_date_end)
        return [rule.date() for rule in date_rules]

    def get_back_scrape_parameters(self, d):
        return {
            'c_page_size': '50',
            'c__ff_cms_opinions_case_name_operator': 'like',
            'c__ff_cms_opinions_case_num_operator': 'like',
            'c__ff_cms_opinions_case_origin_operator': 'like',
            'c__ff_cms_opinions_case_origin': 'like',
            'c__ff_j1_name_operator': 'like%25',
            'c__ff_j2_name_operator': 'like%25',
            'c__ff_cms_opinions_case_type_operator': '%3D',
            'c__ff_cms_opinion_date_published_operator': 'like',
            'c__ff_cms_opinion_date_published': d.strftime('%m/%d/%Y'),
            'c__ff_onSUBMIT_FILTER': 'Search'
        }

    def get_base_path(self):
        return ('//table[@id="search-results-table"]//tr['
                '    not(@id="c_row_") and '
                '    not('
                '        contains(child::td//text(), "NO OPINIONS") or'
                '        contains(child::td//text(), "No Opinions") or'
                '        contains(child::td//text(), "NO MEMO") or'
                '        contains(child::td//text(), "No Memo")'
                '    )'
                ']' + self.position)

    def _get_case_names(self):
        path = '%s/td[1]/a/text()' % self.base
        return [titlecase(text) for text in self.html.xpath(path)]

    def _get_download_urls(self):
        path = '%s/td[1]/a/@href' % self.base
        return self.html.xpath(path)

    def _get_case_dates(self):
        path = '%s/td[7]//text()' % self.base
        return [convert_date_string(date_string) for date_string in self.html.xpath(path)]

    def _get_docket_numbers(self):
        path = '%s/td[2]' % self.base
        return [cell.text_content() for cell in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        if 'opinion' in self.url.lower():
            status = 'Published'
        elif 'memoranda' in self.url.lower():
            status = 'Unpublished'
        else:
            status = 'Unknown'
        return [status] * len(self.case_names)

    def _get_nature_of_suit(self):
        natures = []
        path = '%s/td[5]' % self.base
        for cell in self.html.xpath(path):
            text = cell.text_content()
            natures.append('' if text.lower().strip() == 'n/a' else text)
        return natures

    def _get_lower_court(self):
        path = '%s/td[3]//text()' % self.base
        return self.html.xpath(path)

    def _download_backwards(self, d):
        self.method = 'POST'
        self.url = self.back_scrape_url
        self.parameters = self.get_back_scrape_parameters(d)

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.courtswv.gov/supreme-court/opinions.html'
        self.court_id = self.__module__
        self.cell_path = '//table/tbody/tr/td[%d]'

    def _get_case_names(self):
        return [anchor.text_content() for anchor in self.html.xpath('%s/a[1]' % (self.cell_path % 3))]

    def _get_download_urls(self):
        return [href for href in self.html.xpath('%s/a[1]/@href' % (self.cell_path % 3))]

    def _get_case_dates(self):
        return [convert_date_string(cell.text_content().strip()) for cell in self.html.xpath(self.cell_path % 1)]

    def _get_docket_numbers(self):
        return [cell.text_content().lower().strip() for cell in self.html.xpath(self.cell_path % 2)]

    def _get_precedential_statuses(self):
        codes = {'MD': 'Published',
                 'SO': 'Published',
                 'PC': 'Published',
                 'SEP': 'Separate', }
        return self.decode_cell_text(self.cell_path % 5, codes)

    def _get_nature_of_suit(self):
        # List is sourced from JS in scraped HTML
        codes = {'CR-F': 'Felony (non-Death Penalty)',
                 'CR-M': 'Misdemeanor',
                 'CR-O': 'Criminal-Other',
                 'TCR': 'Tort, Contract, and Real Property',
                 'PR': 'Probate',
                 'FAM': 'Family',
                 'JUV': 'Juvenile',
                 'CIV-O': 'Civil-Other',
                 'WC': 'Workers Compensation',
                 'TAX': 'Revenue (Tax)',
                 'ADM': 'Administrative Agency-Other',
                 'MISC': 'Appeal by Right-Other',
                 'OJ-H': 'Habeas Corpus',
                 'OJ-M': 'Writ Application-Other',
                 'OJ-P': 'Writ Application-Other',
                 'L-ADM': 'Bar Admis   sion',
                 'L-DISC': 'Bar Discipline/Eligibility',
                 'L-DISC-O': 'Bar/Judiciary Proceeding-Other',
                 'J-DISC': 'Bar/Judiciary Proceeding-Other',
                 'CERQ': 'Certified Question',
                 'OJ-O': 'Original Proceeding/Appellate Matter-Other',
                 'POST': 'Post-Conviction Appeal', }
        return self.decode_cell_text(self.cell_path % 4, codes)

    def decode_cell_text(self, path, codes):
        results = []
        for cell in self.html.xpath(path):
            code = cell.text_content().strip()
            results.append(codes[code] if code in codes else 'Unknown')
        return results



from six.moves.urllib.parse import urlencode
from datetime import date, timedelta, datetime

import re
from dateutil.rrule import DAILY, rrule
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.base_url = 'http://media.ca1.uscourts.gov/cgi-bin/opinions.pl'
        self.court_id = self.__module__
        today = date.today()
        params = urlencode({
            'FROMDATE': (today - timedelta(7)).strftime('%m/%d/%Y'),
            'TODATE': today.strftime('%m/%d/%Y'),
            'puid': ''
        })
        self.url = "{}/?{}".format(self.base_url, params)
        # self.url = "http://media.ca1.uscourts.gov/cgi-bin/opinions.pl/?TODATE=06%2F24%2F1993&puid=&FROMDATE=05%2F25%2F1993"
        self.interval = 30
        self.back_scrape_iterable = [i.date() for i in rrule(
            DAILY,
            interval=self.interval,
            dtstart=date(1992, 1, 1),
            until=date(2016, 1, 1),
        )]

    def _get_case_names(self):
        return [e.strip() for e in
                self.html.xpath("//tr[position() > 1]/td[4]/text()[contains(., 'v.')]")]

    def _get_download_urls(self):
        return [e for e in
                self.html.xpath('//tr[position() > 1]/td[2]//@href')]

    def _get_case_dates(self):
        dates = []
        for s in self.html.xpath('//tr[position() > 1]/td[1]//text()'):
            s = s.replace(r'\t', '').replace(r'\n', '').strip()
            if s == '1996/05/32':
                s = '1996/05/30'  # My life is thus lain to waste.
            dates.append(datetime.strptime(s.strip(), '%Y/%m/%d').date())
        return dates

    def _get_docket_numbers(self):
        regex = re.compile(r"(\d{2}-.*?\W)(.*)$")
        docket_numbers = []
        for s in self.html.xpath('//tr[position() > 1]/td[2]/a/text()'):
            s = s.replace('O1-', '01-')  # I grow older, the input grows worse.
            docket_numbers.append(
                regex.search(s).group(1).strip().replace('.', '')
            )
        return docket_numbers

    def _get_precedential_statuses(self):
        statuses = []
        for text in self.html.xpath('//tr[position() > 1]/td[2]//@href'):
            if "U" in text:
                statuses.append("Unpublished")
            elif "P" in text:
                statuses.append("Published")
            elif "E" in text:
                statuses.append("Errata")
            else:
                statuses.append("Unknown")
        return statuses

    def _get_lower_courts(self):
        lower_courts = []
        for e in self.html.xpath('//tr[position() > 1]/td[4]/font'):
            try:
                lower_courts.append(e.xpath('./text()')[0].strip())
            except IndexError:
                lower_courts.append('')
        return lower_courts

    def _download_backwards(self, d):
        params = urlencode({
            'FROMDATE': d.strftime('%m/%d/%Y'),
            'TODATE': (d + timedelta(self.interval)).strftime('%m/%d/%Y'),
            'puid': ''
        })
        self.url = "{}/?{}".format(self.base_url, params)

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200

    def _post_parse(self):
        """This will remove the cases without a case name"""
        to_be_removed = [index for index, case_name in
                         enumerate(self.case_names)
                         if not case_name.replace('v.', '').strip()]

        for attr in self._all_attrs:
            item = getattr(self, attr)
            if item is not None:
                new_item = self.remove_elements(item, to_be_removed)
                self.__setattr__(attr, new_item)

    @staticmethod
    def remove_elements(list_, indexes_to_be_removed):
        return [i for j, i in enumerate(list_) if j not in
                indexes_to_be_removed]



"""Scraper for Second Circuit
CourtID: ca2
Contact:
  Webmaster@ca2.uscourts.gov ('Calendar Team' handles updates, and is responsive)
  ecfhelpdesk@ca2.uscourts.gov
"""


from juriscraper.opinions.united_states.federal_appellate import ca2_p


class Site(ca2_p.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = "http://www.ca2.uscourts.gov/decisions?IW_DATABASE=SUM&IW_FIELD_TEXT=*&IW_SORT=-Date&IW_BATCHSIZE=100"
        self.court_id = self.__module__



from juriscraper.opinions.united_states.federal_special import cit

class Site(cit.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.cit.uscourts.gov/SlipOpinions/SlipOps-2011.html'
        self.court_id = self.__module__




"""Scraper for the Supreme Court of Delaware
CourtID: del

Creator: Andrei Chelaru
Reviewer: mlr
"""

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://courts.delaware.gov/opinions/list.aspx?ag=supreme%20court'
        # Note that we can't do the usual thing here because 'del' is a Python keyword.
        self.court_id = 'juriscraper.opinions.united_states.state.del'

    def _get_case_dates(self):
        return [convert_date_string(date.strip()) for date in self._get_nth_cell_data(2, text=True)]

    def _get_download_urls(self):
        return [url.strip() for url in self._get_nth_cell_data(1, href=True)]

    def _get_case_names(self):
        return [name.strip() for name in self._get_nth_cell_data(1, link_text=True)]

    def _get_docket_numbers(self):
        return [docket.strip() for docket in self._get_nth_cell_data(3, link_text=True)]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_dates)

    def _get_judges(self):
        # We need special logic here because they use <br> tags in the cell text
        return [' '.join(cell.xpath('text()')) for cell in self.html.xpath('//table/tr/td[6]')]

    def _get_nth_cell_data(self, cell_number, text=False, href=False, link_text=False):
        # Retrieve specific types of data from all nTH cells in the table
        path = '//table/tr/td[%d]' % cell_number
        if text:
            path += '/text()'
        elif href:
            path += '/a/@href'
        elif link_text:
            path += '/a/text()'
        return self.html.xpath(path)



"""
Scraper for the United States Bankruptcy Appellate Panel for the Tenth Circuit
CourtID: bap10
Court Short Name: 10th Cir. BAP
Auth: Jon Andersen <janderse@gmail.com>
Reviewer: mlr
History:
    2014-09-01: First draft by Jon Andersen
    2014-09-02: Revised by mlr to use _clean_text() instead of pushing logic
                into the _get_case_dates function.
"""

from datetime import datetime
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.bap10.uscourts.gov/opinions/new/opinion.txt'
        self.court_id = self.__module__
        self.my_case_names = []
        self.my_download_urls = []
        self.my_docket_numbers = []
        # Site has non-chained, bad certificate, need to
        # ignore ssl verification for now for scraper to work
        self.request['verify'] = False

    def _clean_text(self, text):
        """This page is a txt file, so here we convert it to something that
        can be easily ingested by lxml.

        Lines like:
          13-94.pdf|13|94|08/25/2014|Steve Christensen|Raymond Madsen|United States Bankruptcy Court for the District of Utah
        """
        # Nuke duplicates
        lines = set([line for line in text.split('\n') if line])

        # Build an XML tree.
        xml_text = '<rows>\n'
        for line in lines:
            values = line.split('|')
            xml_text += '  <row>\n'
            for value in values:
                xml_text += '    <value>%s</value>\n' % value
            xml_text += '  </row>\n'
        xml_text += '</rows>\n'
        return xml_text

    def _get_case_dates(self):
        path = '//row/value[4]/text()'
        return [datetime.strptime(date_string, '%m/%d/%Y').date()
                for date_string in self.html.xpath(path)]

    def _get_case_names(self):
        plaintiffs = self.html.xpath('//row/value[6]/text()')
        defendants = self.html.xpath('//row/value[7]/text()')
        case_names = []
        for p, d in zip(plaintiffs, defendants):
            case_names.append("%s v. %s" % (p, d))
        return case_names

    def _get_download_urls(self):
        years = self.html.xpath('//row/value[2]/text()')
        file_names = self.html.xpath('//row/value[1]/text()')
        urls = []
        for y, f_n in zip(years, file_names):
            urls.append('http://www.bap10.uscourts.gov/opinions/%s/%s' % (y, f_n))
        return urls

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_dates)

    def _get_docket_numbers(self):
        years = self.html.xpath('//row/value[2]/text()')
        numbers = self.html.xpath('//row/value[3]/text()')
        docket_numbers = []
        for y, n in zip(years, numbers):
            docket_numbers.append("%s-%s" % (y, n))
        return docket_numbers



# Scraper for New York Appellate Divisions 1st Dept.
#CourtID: nyappdiv_1st
#Court Short Name: NY
#Author: Andrei Chelaru
#Reviewer: mlr
#Date: 2014-07-04

from datetime import date
from dateutil.rrule import rrule, MONTHLY

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.html_utils import get_html5_parsed_text
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        date_keys = rrule(MONTHLY, dtstart=date(2003, 11, 1), until=date(2015, 8, 30))
        self.back_scrape_iterable = [i.date() for i in date_keys]
        self.row_base_path = '//tr[contains(./td[1]/a/@href, "3d")]'
        self.division = 1
        self.url = self.build_url()

    def _get_case_names(self):
        path = '%s/td[1]' % self.row_base_path
        return [cell.text_content() for cell in self.html.xpath(path)]

    def build_url(self, target_date=False):
        base = 'http://www.courts.state.ny.us/reporter/slipidx/aidxtable_%s' % self.division
        if target_date:
            return '%s_%s_%s.shtml' % (base, target_date.year, target_date.strftime("%B"))
        else:
            return '%s.shtml' % base

    def _get_download_urls(self):
        path = '%s/td[1]//a/@href' % self.row_base_path
        return self.html.xpath(path)

    def _get_case_dates(self):
        case_dates = []
        for element in self.html.xpath('//caption | //center'):
            date_string = element.text_content().strip().replace('Cases Decided ', '')
            path_prefix = './parent::' if element.tag == 'caption' else './following-sibling::'
            path = path_prefix + 'table[1]' + self.row_base_path
            cases = element.xpath(path)
            case_dates.extend([convert_date_string(date_string)] * len(cases))
        return case_dates

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        path = '%s/td[3]' % self.row_base_path
        return map(self._add_str_to_list_where_empty_element, self.html.xpath(path))

    def _get_judges(self):
        path = '%s/td[2]' % self.row_base_path
        return map(self._add_str_to_list_where_empty_element, self.html.xpath(path))

    def _get_neutral_citations(self):
        path = '%s/td[4]' % self.row_base_path
        return [cell.text_content().strip() for cell in self.html.xpath(path)]

    @staticmethod
    def _add_str_to_list_where_empty_element(element):
        string_list = element.xpath('./text()')
        return string_list[0] if string_list else ''

    def _download_backwards(self, target_date):
        self.crawl_date = target_date
        logger.info("Running backscraper with date: %s" % target_date)
        self.url = self.build_url(target_date=target_date)
        self.html = self._download()

    def _make_html_tree(self, text):
        return get_html5_parsed_text(text)



"""Scraper for Army Court of Criminal Appeals
CourtID: acca
Reviewer: None
History:
  2015-01-08: Created by mlr
  2016-03-17: Website appears to be dead. Scraper disabled in __init__.py.
"""


import re
from lxml import html

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.jagcnet.army.mil/85257546006DF36B/ODD?OpenView&Count=-1'
        self.docket_case_name_splitter = re.compile('(.*[\dX]{5,8})(.*)')
        # DoD uses an annoying certificate that isn't installed. Thus, we
        # disable certificate verification in requests.
        self.request['verify'] = False

    def _get_download_urls(self):
        path = "//a[@class='td_frmOpinion']/@href"
        return list(self.html.xpath(path))

    def _get_case_names(self):
        path = "//a[@class='td_frmOpinion']"
        case_names = []
        for e in self.html.xpath(path):
            s = html.tostring(e, method='text', encoding='unicode')
            m = self.docket_case_name_splitter.search(s)
            case_names.append(m.group(2))
        return case_names

    def _get_case_dates(self):
        path = "//a[@class='DtOpn2']/text()"
        return [convert_date_string(date_string) for date_string in  self.html.xpath(path)]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        path = "//a[@class='td_frmOpinion']"
        docket_numbers = []
        for e in self.html.xpath(path):
            s = html.tostring(e, method='text', encoding='unicode')
            m = self.docket_case_name_splitter.search(s)
            docket_numbers.append(m.group(1))
        return docket_numbers



"""Scraper for U.S. Merit Systems Protection Board
CourtID: mspb
Court Short Name: MSPB
Author: Jon Andersen
Reviewer: mlr
Date created: 1 Sep 2014
Type: Non-precedential
"""

from juriscraper.opinions.united_states.administrative_agency import mspb_p


class Site(mspb_p.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.type = 'Nonprecedential'
        self.display = 60414
        self.column_diff = -1
        self.set_url()

    def _get_precedential_statuses(self):
        return ["Unpublished"] * len(self.case_dates)

    def _get_neutral_citations(self):
        return None



# coding=utf-8
import re
import urlparse

from juriscraper.pacer.reports import BaseReport
from .docket_report import BaseDocketReport
from .utils import clean_pacer_object
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import convert_date_string, force_unicode, harmonize

logger = make_default_logger()


class ClaimsRegister(BaseDocketReport, BaseReport):
    """Query and parse the claims registry.

    This is the most naive version of this parser possible. It does not support
    *any* variation to the table, and assumes that the table is *always*
    queried in the same way, using the default options. Because this table is
    not uploaded by RECAP, we do not need to support all the possible versions,
    only those that we need for a particular client job. Enhancements welcome
    to make it more flexible on the parsing and featureful on the querying.
    """
    PATH = 'cgi-bin/SearchClaims.pl'
    CACHE_ATTRS = ['claims', 'metadata']

    def __init__(self, court_id, pacer_session=None):
        super(ClaimsRegister, self).__init__(court_id, pacer_session)
        self._clear_caches()
        self._metadata = None
        self._claims = None
        assert court_id.endswith('b'), \
            "Unable to create object. Must use bankruptcy court abbreviation."

    @property
    def data(self):
        """Get all the data back from this endpoint."""
        if self.is_valid is False:
            return {}
        data = self.metadata.copy()
        data['claims'] = self.claims
        return data

    @property
    def metadata(self):
        if self._metadata is not None:
            return self._metadata

        data = {
            'court_id': self.court_id,
        }
        field_mappings = {
            'case_name': 'case_name',
            'case_number': 'docket_number',
            'chapter': 'chapter',
            'closed': 'date_closed',
            'converted': 'date_converted',
            'date_filed': 'date_filed',
            'debtor_dismissed': 'date_debtor_dismissed',
            'judge': 'assigned_to_str',
            'office': 'office',
            'trustee': 'trustee_str',
            'last_date_to_file_claims': 'date_last_to_file_claims',
            'last_date_to_file_govt': 'date_last_to_file_govt',
        }
        # The title row has some weird labels that look like:
        #  <span class="label">Closed</span> 10/26/2018
        label_nodes = self.tree.xpath('//span[@class="label"]')
        for label_node in label_nodes:
            data.update(self._get_label_value_pair(label_node, False,
                                                   field_mappings))

        # Each cell in the header table has a value like:
        # <td><b>Judge:</b>  Barbara J. Houser</td>
        cells = self.tree.xpath('//center/p/table[1]//td[text()]')
        for cell in cells:
            label_node = cell.xpath('./b')[0]
            data.update(self._get_label_value_pair(label_node, True,
                                                   field_mappings))

        # Finally, do the cells in the footer area. These look like:
        # <b>Case Name:</b> Scottish Holdings, Inc. <br>
        label_nodes = self.tree.xpath('//center[1]/b')
        for label_node in label_nodes:
            data.update(self._get_label_value_pair(label_node, True,
                                                   field_mappings))

        data['docket_number'] = self._clean_docket_number(
            data.get('docket_number', ''))

        data['case_name'] = harmonize(data['case_name'])
        data = clean_pacer_object(data)
        self._metadata = data
        return data

    def _clean_docket_number(self, docket_number):
        """Strip the judge initials off docket numbers, and do any other
        similar cleanup.

        :param: docket_number: A string docket number to clean.
        """
        # Uses both b/c sometimes the bankr. cases have a dist-style docket
        # number.
        regexes = [self.docket_number_dist_regex,
                   self.docket_number_bankr_regex]
        for regex in regexes:
            match = regex.search(docket_number)
            if match:
                return match.group(1)

        # Fail safe
        return docket_number

    @property
    def claims(self):
        """Parse out all the claims information"""
        if self._claims is not None:
            return self._claims

        claim_tables = self.tree.xpath('//table[@class="complexReport"]')
        claims = []
        for claim_table in claim_tables:
            claim = {}

            # Identify and parse each cell.
            creditor_td = claim_table.xpath('(.//tr)[1]/td[1]')[0]
            claim.update(self._parse_creditor_cell(creditor_td))

            claim_number_td = claim_table.xpath('(.//tr)[1]/td[2]')[0]
            claim.update(self._parse_claim_number_cell(claim_number_td))

            claim_metadata_td = claim_table.xpath('(.//tr)[1]/td[3]')[0]
            claim.update(self._parse_claim_metadata_cell(claim_metadata_td))

            amounts_td = claim_table.xpath('(.//tr)[2]/td[1]')[0]
            claim.update(self._parse_amounts_cell(amounts_td))

            # Paths here are a bit more complex because the amounts table above
            # contains a table, so we can't just get trs by index anymore.
            history_path = './/tr[.//i[text() = "History:"]]'
            history_td = claim_table.xpath(history_path)[0]
            claim['history'] = self._parse_history_cell(history_td)

            description_path = './/i[text() = "Description:"]'
            description_node = claim_table.xpath(description_path)[0]
            claim.update(self._get_label_value_pair(
                description_node, True, {'Description': 'description'}))

            remarks_path = './/i[text() = "Remarks:"]'
            remarks_td = claim_table.xpath(remarks_path)[0]
            claim.update(self._get_label_value_pair(
                remarks_td, True, {'remarks': 'remarks'}))

            claims.append(claim)

        self._claims = claims
        return claims

    def _parse_creditor_cell(self, td):
        """Get creditor_id, creditor_details"""
        credit_label = td.xpath('.//i')[0]
        info = self._get_label_value_pair(credit_label, True,
                                          {'creditor': 'creditor_id'})
        info['creditor_id'] = info['creditor_id'].strip('()')

        # Redelimit the <br> tags, then use newlines to join everything but the
        # first line in the cell, and the "Claimant History" link if applicable
        better_td = self.redelimit_p(td, self.BR_REGEX)
        good_text = [s.strip() for s in better_td.xpath('.//p/text()')[1:]]
        info['creditor_details'] = '\n'.join(good_text).strip()

        return info

    def _parse_claim_number_cell(self, td):
        """Get the claim_number and date fields"""
        data = {}
        claim_number_text = td.xpath('.//b')[0].text_content()
        claim_number = int(re.search('\d+', claim_number_text).group(0))
        data['claim_number'] = claim_number

        labels = td.xpath('.//i')
        fields = {
            'original_filed_date': 'date_original_filed',
            'original_entered_date': 'date_original_entered',
            'last_amendment_filed': 'date_last_amendment_filed',
            'last_amendment_entered': 'date_last_amendment_entered',
        }
        for label in labels:
            data.update(self._get_label_value_pair(label, False, fields))
        return data

    def _parse_claim_metadata_cell(self, td):
        """Get the status, filed_by, entered_by, date_modified fields"""
        labels = td.xpath('.//i')
        data = {}
        fields = {
            'modified': 'date_modified',
        }
        for label in labels:
            data.update(self._get_label_value_pair(label, True, fields))
        return data

    @staticmethod
    def _parse_amounts_cell(td):
        """Get the amounts claimed/secured/etc."""
        # These fields are tricky. They come in a table like this:
        #  | Amount  | Claimed: | $49.00
        #  | Secured | Claimed: | $49.00
        # We go row by row.
        fields = {
            'Amount': 'amount_claimed',
            'Secured': 'secured_claimed',
            'Priority': 'priority_claimed',
            'Unsecured': 'unsecured_claimed',
        }
        rows = td.xpath('.//tr')
        data = {}
        for row in rows:
            # The first cell sets the field name
            label = row.xpath('.//td[1]')[0].text_content().strip()
            if not label:
                # There's an empty row; ignore it.
                continue
            label = fields[label]
            value = row.xpath('.//td[3]')[0].text_content().strip()
            data[label] = value
        return data

    @staticmethod
    def _parse_history_cell(td):
        """Parse the history table.

        This is similar to a mini-docket entries table, but it has a mixture of
        claims documents (with their own numbering scheme), and normal docket
        entries.
        """
        history_rows = []
        entry_trs = td.xpath('.//tr')
        for entry_tr in entry_trs:
            row = {}
            number_cell = entry_tr.xpath('./td[3]')[0]
            number = number_cell.text_content().strip()
            if '-' in number:
                row['type'] = 'claim'
                nums = [int(v) for v in number.split('-')]
                row['document_number'] = nums[0]
                row['attachment_number'] = nums[1]
            else:
                row['type'] = 'docket_entry'
                try:
                    row['document_number'] = int(number)
                except ValueError:
                    # Minute entry
                    row['document_number'] = None

            # Next do the URL parsing for whatever is in it (if we have a URL)
            if row['document_number'] is not None:
                href = number_cell.xpath('.//a/@href')[0]
                qs = href.rsplit('?')[1]
                qs_dict = urlparse.parse_qs(qs)
                if row['type'] == 'claim':
                    row['id'] = qs_dict['claim_id'][0]
                    row['pacer_case_id'] = qs_dict['caseid'][0]
                elif row['type'] == 'docket_entry':
                    # Unfortunately, no doc1 value. There is a dm_id, which I'm
                    # not familiar with, but I'm not sure what it does, and we
                    # lack a field for it in our DB.
                    row['pacer_seq_no'] = qs_dict['de_seq_num'][0]
                    row['pacer_case_id'] = qs_dict['caseid'][0]
                    row['pacer_dm_id'] = qs_dict['dm_id'][0]

            # Date
            date_cell = entry_tr.xpath('./td[4]')[0]
            row['date_filed'] = convert_date_string(date_cell.text_content())

            # Description
            desc_cell = entry_tr.xpath('./td[5]')[0]
            row['description'] = force_unicode(
                desc_cell.text_content().strip())
            history_rows.append(row)
        return history_rows

    def query(self, pacer_case_id, date_start=None, date_end=None):
        """Query the claims register and return the results.

        :param pacer_case_id: The internal PACER ID for the case
        :type pacer_case_id: str
        :param date_start: The earliest claim entry you wish to have. Default
        is 1960-01-01 (before anything was put in PACER)
        :type date_start: date
        :param date_end: The latest claim entry you with to have. Default is
        tomorrow.
        :type date_end: date
        :return: request response object
        """
        assert self.session is not None, \
            "session attribute of ClaimsRegister cannot be None."

        params = {
            'all_case_ids': pacer_case_id,
            # Whether to filter by filed date or entered date. Parser only
            # currently supports filed date.
            'EntFil': 'Fil',
            'SearchingClaims': 'true',
            'f_sort1': 'cl_claimno',
            'f_sort2': 'cl_dt_filed',
            # This field seems to be required while 'all_case_ids' is not.
            'case_num': '2:17-bk-31853',
        }
        if date_start:
            params['f_fromdt'] = date_start.strftime(u'%m/%d/%Y')
        else:
            # Set the default
            params['f_fromdt'] = '1/1/1960'
        if date_end:
            params['f_todt'] = date_end.strftime(u'%m/%d/%Y')

        logger.info("Querying claims register for case ID '%s' with "
                    "params %s", pacer_case_id, params)
        self.response = self.session.post(self.url + '?1-L_1_0-1',
                                          data=params)
        self.parse()



from .string_utils import titlecase

titles = [
    'judge', 'magistrate', 'district', 'chief', 'senior',
    'bankruptcy', 'mag.', 'magistrate-judge', 'mag/judge', 'mag',
    'visiting', 'special', 'senior-judge', 'master', 'u.s.magistrate',
]
blacklist = [
    'a998', 'agb', 'am', 'associated', 'ca', 'cases', 'cet', 'ch.', 'cla',
    'clerk', 'cp', 'cvb', 'db', 'debt-magistrate', 'discovery', 'dj', 'docket',
    'duty', 'duty', 'ec', 'eck', 'general', 'grc', 'gs', 'hhl', 'hon',
    'honorable', 'inactive', 'jne', 'jv', 'kec', 'law', 'lc', 'llh', 'lq',
    'maryland', 'mediator', 'merged', 'mj', 'mmh', 'msh', 'mwd', 'no', 'none',
    'prisoner', 'pslc', 'pro', 'pso', 'pt', 'rmh', 'se', 'sf', 'show',
    'successor', 'u.s.', 'tjc', 'unassigned', 'unassigned2', 'unassigneddj',
    'unknown', 'us', 'usdc', 'vjdistrict',
]

judge_normalizers = {
    # Generic Judge & Bankruptcy Judge we merge together for simplicity.
    '': 'jud',
    'Judge': 'jud',
    'Judge Judge': 'jud',
    'District Judge': 'jud',
    'Visiting Judge': 'jud',
    'Bankruptcy': 'jud',
    'Bankruptcy Judge': 'jud',

    # Chief (normalize to jud for now, due to low sample size)
    'Chief': 'jud',
    'Chief Judge': 'jud',
    'Chief District Judge': 'jud',

    # Magistrate
    'Mag': 'mag',
    'Mag Judge': 'mag',
    'mag/judge': 'mag',
    'Magistrate': 'mag',
    'Magistrate Judge': 'mag',
    'Magistrate-Judge': 'mag',
    'Magistrate Judge Mag': 'mag',
    'Magistrate Judge Magistrate': 'mag',
    'Magistrate Judge Magistrate Judge': 'mag',

    # Chief Magistrate (normalize to magistrate for now, due to low sample size)
    'Chief Magistrate': 'mag',
    'Chief Magistrate Judge': 'mag',

    # Senior
    'Senior Judge': 'ret-senior-jud',
    'Senior-Judge': 'ret-senior-jud',

    # Special Master
    'Special Master': 'spec-m',
    'Chief Special Master': 'c-spec-m',
}


def normalize_judge_titles(title):
    """Normalize judge titles

    Take in a string like "Magistrate Judge" and return the normalized
    abbreviation from the POSITION_TYPES variable. Assumes that input is
    titlecased.

    Also normalizes things like:
     - District Judge --> Judge
     - Blank --> Judge
     - Bankruptcy Judge --> Judge
    """
    return judge_normalizers.get(title, 'UNKNOWN: %s' % title)


def normalize_judge_names(name):
    """Cleans up little issues with people's names"""
    out = []
    words = name.split()
    for i, w in enumerate(words):
        # Michael J Lissner --> Michael J. Lissner
        if len(w) == 1 and w.isalpha():
            w = '%s.' % w

        # Michael Lissner Jr --> Michael Lissner Jr.
        if w.lower() in ['jr', 'sr']:
            w = '%s.' % w

        # J. Michael Lissner --> Michael Lissner
        # J. G. Lissner --> J. G. Lissner
        if i == 0 and w.lower() in ['j.', 'j']:
            next_word = words[i + 1]
            if not any([len(next_word) == 2 and next_word.endswith('.'),
                        len(next_word) == 1]):
                # Drop the word.
                continue
        out.append(w)

    return ' '.join(out)


def normalize_judge_string(judge):
    """Split a string representing a judge returning their name and title.

    This code was generated and tested against all judge strings in the RECAP
    Archive.

    :param judge: A string representing a judge, e.g.
    :return: A tuple of the judge's name and their normalized position.

    >>> normalize_judge_string('Honorable Sue W. Wright')
    ('Sue W. Wright', 'jud')
    """
    judge = judge.replace(',', '')
    words = judge.lower().split()

    # Nuke bad junk (punct., j, blacklist, etc.)
    title_words = []
    name_words = []
    for w in words:
        contains_parens = '(' in w or ')' in w
        starts_poorly = w.startswith('-') or w.startswith('~')
        blacklisted = w in blacklist
        long_abbrev = (len(w) > 2 and '.' in w and w not in ['jr.', 'sr.'])
        if any([contains_parens, starts_poorly, blacklisted, long_abbrev]):
            continue
        if w in titles:
            title_words.append(w)
        else:
            name_words.append(w)

    title = normalize_judge_titles(titlecase(' '.join(title_words)))
    name = normalize_judge_names(titlecase(' '.join(name_words)))

    return name, title



import time
from datetime import date
from lxml import html
from dateutil.rrule import DAILY, rrule

from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        today = date.today()
        self.url = 'http://www.ca10.uscourts.gov/opinion/search/results?query=%20date%3A{}'.format(
            today.strftime('%m/%d/%Y')
        )
        self.court_id = self.__module__
        self.back_scrape_iterable = [i.date() for i in rrule(
            DAILY,
            dtstart=date(1995, 11, 1),
            until=date(2015, 1, 1),
        )]
        self.base = "//table[contains(concat('', @class, ''), 'search-results')]//tr"

    def _get_case_names(self):
        return [e for e in self.html.xpath("{}/td[@class='case-name']/text()".format(self.base))]

    def _get_download_urls(self):
        return [e for e in self.html.xpath("{}/td[@class='link']/a/@href".format(self.base))]

    def _get_case_dates(self):
        dates = []
        for date_string in self.html.xpath("{}/td[@class='publish-date']/text()".format(self.base)):
            # ex: Nov-02-1995
            dates.append(date.fromtimestamp(time.mktime(time.strptime(date_string, '%b-%d-%Y'))))
        return dates

    def _get_docket_numbers(self):
        return [e for e in self.html.xpath("{}/td[@class='case-no']/text()".format(self.base))]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_lower_courts(self):
        return [e for e in self.html.xpath("{}/td[@class='origin']/text()".format(self.base))]

    def _download_backwards(self, d):

        self.url = 'http://www.ca10.uscourts.gov/opinion/search/results?query=%20date%3A{}'.format(
            d.strftime('%m/%d/%Y')
        )

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



"""Scraper for Vermont Supreme Court
CourtID: vt
Court Short Name: VT
Court Contact: submit form here https://www.vermontjudiciary.org/website-feedback-form

If there are errors with the site, you can contact:

 Monica Bombard
 (802) 828-4784

She's very responsive.
"""

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.back_scrape_iterable = range(1, self.get_backscrape_max())
        self.element_path_format = "//article[@class='views-row media-document']/div[@class='views-field views-field-field-document%s']"
        self.url = 'https://www.vermontjudiciary.org/opinions-decisions?f[0]=document_type%3A94&f[1]=court_division_opinions_library_%3A' + self.get_division_id()
        self.backscrape_page_base_url = self.url + '&page='

    def get_backscrape_max(self):
        return 7

    def get_division_id(self):
        return '7'

    def _get_download_urls(self):
        path_base = self.element_path_format % ''
        path = "%s//@href" % path_base
        return self.html.xpath(path)

    def _get_case_names(self):
        path = self.element_path_format % ''
        return [e.text_content().strip() for e in self.html.xpath(path)]

    def _get_case_dates(self):
        path = self.element_path_format % '-expiration'
        return [convert_date_string(e.text_content()) for e in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        path = self.element_path_format % '-number'
        return [e.text_content().strip() for e in self.html.xpath(path)]

    def _download_backwards(self, page_number):
        logger.info("PROCESSING PAGE: %d" % (page_number + 1))
        self.url = self.backscrape_page_base_url + str(page_number)
        self.html = self._download()


# History:
# - Long ago: Created by mlr
# - 2014-11-07: Updated by mlr to use new website.

from datetime import datetime

from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.ca5.uscourts.gov/rss.aspx?Feed=Opinions&Which=All&Style=Detail'
        self.court_id = self.__module__

    def _get_case_names(self):
        path = '//item/description/text()[2]'
        return [s for s in self.html.xpath(path)]

    def _get_download_urls(self):
        path = '//item/link'
        return [e.tail for e in self.html.xpath(path)]

    def _get_case_dates(self):
        path = '//item/description/text()[5]'
        return [datetime.strptime(date_string, '%m/%d/%Y').date()
                for date_string in self.html.xpath(path)]

    def _get_docket_numbers(self):
        path = '//item/description/text()[1]'
        return [s for s in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        path = '//item/description/text()[3]'
        statuses = []
        for status in self.html.xpath(path):
            if status == 'pub':
                statuses.append('Published')
            elif status == 'unpub':
                statuses.append('Unpublished')
            else:
                statuses.append('Unknown')
        return statuses

    def _get_nature_of_suit(self):
        path = '//item/description/text()[4]'
        return [s for s in self.html.xpath(path)]



"""
Auth: Jordan Atanasov <jordan.atanasov@commetric.com>
History:
    2012-05-07: Written by Jordan.
    2012-07-06: Updated by mlr to only get the first ten items.
    2015-07-28: Updated by m4h7 to have improved xpaths.
        Notes: Only queries first ten dates. Beyond that, they get messy.
    2017-01-11: Updated by arderyp to not suppress opinions under
        date range headings.  We'v added not logic to OpinionSite
        to handle estimated dates, and are now flagging these
        examples as such
"""

import datetime
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string, split_date_range_string, normalize_dashes


class Site(OpinionSite):
    url_court_id = 'sp'

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'https://public.courts.alaska.gov/web/appellate/%s.htm' % self.url_court_id
        self.court_id = self.__module__
        self.date_string_path = '//h4'
        self.sub_opinion_path = './following-sibling::ul[1]//a/em'

    def _get_case_names(self):
        path = "//ul/li[descendant::a/em]//em/text()"
        return [t for t in self.html.xpath(path)]

    def _get_download_urls(self):
        path = "//ul/li/a[child::em]/@href"
        return [h for h in self.html.xpath(path)]

    def _get_case_dates(self):
        dates = []
        for element in self.html.xpath(self.date_string_path):
            # Determine the number of opinions below the
            # date and add them all to the date list.
            count = len(element.xpath(self.sub_opinion_path))
            # Handle different date formats
            date_string = element.text_content().strip()
            try:
                # Normal date
                date = convert_date_string(date_string)
            except ValueError:
                # It a date range string like 'January - March 2016'
                # return the middle date, unless its in the future
                # in which case return today's date
                middle_date = split_date_range_string(date_string)
                date = min(datetime.date.today(), middle_date)
            dates.extend([date] * count)
        return dates

    def _get_docket_numbers(self):
        path = "//ul/li[descendant::a/em]/text()[1]"
        return [t for t in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_date_filed_is_approximate(self):
        approximations = []
        for element in self.html.xpath(self.date_string_path):
            count = len(element.xpath(self.sub_opinion_path))
            date_string = normalize_dashes(element.text_content())
            approximation = True if '-' in date_string else False
            approximations.extend([approximation] * count)
        return approximations



# coding=utf-8
import re

import six
from lxml import etree
from lxml.html import HtmlElement
from six.moves.urllib.parse import urljoin

from .utils import is_pdf, make_doc1_url
from ..lib.html_utils import (clean_html, fix_links_in_lxml_tree,
                              get_html5_parsed_text, get_html_parsed_text,
                              set_response_encoding)
from ..lib.log_tools import make_default_logger

logger = make_default_logger()


# Patch the HtmlElement class to add a function that can handle regular
# expressions within XPath queries. See usages throughout AppellateDocketReport.
def re_xpath(self, path):
    return self.xpath(path, namespaces={
        're': 'http://exslt.org/regular-expressions'})
HtmlElement.re_xpath = re_xpath


class BaseReport(object):
    """A base report for working with pages on PACER."""

    REDIRECT_REGEX = re.compile('window\.\s*?location\s*=\s*"(.*)"\s*;')

    # Subclasses should override PATH
    PATH = ''

    # Strings below (and in subclasses) are used to identify HTML that should
    # not be parsed or processed for a variety of reasons. Spaces in the strings
    # below are converted to \s whitespace searches using regexes.
    ERROR_STRINGS = [
        "MetaMask.*web3",
        'console\.log\(".*CloudMask',
        'Drumpf',
    ]

    def __init__(self, court_id, pacer_session=None):
        self.court_id = court_id
        self.session = pacer_session
        self.tree = None
        self.response = None
        self.is_valid = None
        super(BaseReport, self).__init__()

    @property
    def url(self):
        if self.court_id == 'psc':
            return "https://dcecf.psc.uscourts.gov/%s" % self.PATH
        else:
            return "https://ecf.%s.uscourts.gov/%s" % (self.court_id, self.PATH)

    def query(self, *args, **kwargs):
        """Query PACER and set self.response with the response."""
        raise NotImplementedError(".query() must be overridden")

    def parse(self):
        """Parse the data provided in a requests.response object and set
        self.tree to be an lxml etree. In most cases, you won't need to call
        this since it will be automatically called by self.query, if needed.

        :return: None
        """
        self.response.raise_for_status()
        set_response_encoding(self.response)
        self._parse_text(self.response.text)

    def _parse_text(self, text):
        """Parse the HTML as unicode text and set self.tree

        This is a particularly critical method when running tests, which pull
        from local disk instead of from a query response. This is also used
        when data comes from a source other than self.query() (such as a user
        upload). This method should probably be made public as .parse_text().

        :param text: A unicode object
        :return: None
        """
        assert isinstance(text, six.text_type), \
            "Input must be unicode, not %s" % type(text)
        text = clean_html(text)
        self.check_validity(text)
        if self.is_valid:
            self.tree = get_html5_parsed_text(text)
            etree.strip_elements(self.tree, u'script')
            self.tree.rewrite_links(fix_links_in_lxml_tree, base_href=self.url)

    def check_validity(self, text):
        """Place sanity checks here to make sure that the returned text is
        valid and not an error page or some other kind of problem.

        Set self.is_valid flag to True or False
        """
        for error_string in self.ERROR_STRINGS:
            error_string_re = re.compile('\s+'.join(error_string.split()),
                                         flags=re.I)
            if error_string_re.search(text):
                self.is_valid = False
                return
        self.is_valid = True

    @property
    def data(self):
        """Extract the data from the tree and return it."""
        raise NotImplementedError('.data() must be overridden.')

    def _query_pdf_download(self, pacer_case_id, pacer_doc_id, got_receipt):
        """Query the doc1 download URL.

        :param pacer_case_id: The case ID for the case. Used for disambiguating
        which case the document is associated with when a pacer_doc_id is used
        in multiple cases.
        :param pacer_doc_id: The doc id for the document
        :param got_receipt: Whether to get the receipt for the page ('0') or
        get the PDF itself ('1').
        :return the Request.response object and the url queried
        """
        timeout = (60, 300)
        url = make_doc1_url(self.court_id, pacer_doc_id, True)
        data = {
            'case_id': pacer_case_id,
            'got_receipt': got_receipt,
        }
        logger.info("GETting URL: %s with params: %s" % (url, data))
        r = self.session.get(url, params=data, timeout=timeout)
        return r, url

    def download_pdf(self, pacer_case_id, pacer_doc_id):
        """Download a PDF from PACER.

        Note that this doesn't support attachments yet.

        :returns: request.Response object containing a PDF, if one can be found
        (is not sealed, gone, etc.). Else, returns None.
        """
        r, url = self._query_pdf_download(pacer_case_id, pacer_doc_id,
                                          got_receipt='1')

        if u'This document is not available' in r.text:
            logger.error("Document not available in case: %s at %s", url,
                         pacer_case_id)
            return None

        # Some pacer sites use window.location in their JS, so we have to look
        # for that. See: oknd, 13-cv-00357-JED-FHM, doc #24. But, be warned,
        # you can only catch the redirection with JS off.
        m = self.REDIRECT_REGEX.search(r.text)
        if m is not None:
            r = self.session.get(urljoin(url, m.group(1)))
            r.raise_for_status()

        # The request above sometimes generates an HTML page with an iframe
        # containing the PDF, and other times returns the PDF directly. ∴
        # either get the src of the iframe and download the PDF or just return
        # the pdf.
        r.raise_for_status()
        if is_pdf(r):
            logger.info('Got PDF binary data for case at %s', url)
            return r

        text = clean_html(r.text)
        tree = get_html_parsed_text(text)
        tree.rewrite_links(fix_links_in_lxml_tree,
                           base_href=r.url)
        try:
            iframe_src = tree.xpath('//iframe/@src')[0]
        except IndexError:
            if 'pdf:Producer' in text:
                logger.error("Unable to download PDF. PDF content was placed "
                             "directly in HTML. URL: %s, caseid: %s", url,
                             pacer_case_id)
            else:
                logger.error("Unable to download PDF. PDF not served as "
                             "binary data and unable to find iframe src "
                             "attribute. URL: %s, caseid: %s", url,
                             pacer_case_id)
            return None

        r = self.session.get(iframe_src)
        if is_pdf(r):
            logger.info('Got iframed PDF data for case %s at: %s', url,
                        iframe_src)

        return r

    def is_pdf_sealed(self, pacer_case_id, pacer_doc_id):
        """Check if a PDF is sealed without trying to actually download
        it.
        """
        r, url = self._query_pdf_download(pacer_case_id, pacer_doc_id,
                                          got_receipt='0')
        sealed = 'You do not have permission to view this document.'
        return sealed in r.content




# Scraper for the United States Court of International Trade
# CourtID: cit
# Court Short Name: Ct. Int'l Trade
# Neutral Citation Format: Ct. Int'l Trade No. 12-1
from juriscraper.OpinionSite import OpinionSite
import re
import time
from datetime import date
from lxml import html

class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
# This is a special backscraper to deal with problems on the 2008 page.
        self.url = 'http://www.cit.uscourts.gov/SlipOpinions/SlipOps-2008.html'
        self.court_id = self.__module__

    def _get_download_urls(self):
        return [t for t in self.html.xpath('//table[3]/tr[position() > 1]/td[1]/a/@href | //table[3]/tr[position() > 1]/td[1]/font/a/@href')]

    def _get_neutral_citations(self):
        neutral_citations = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[1]'):
            s = html.tostring(e, method='text', encoding='unicode').strip()
            neutral_citations.append(s)
        return neutral_citations

    def _get_case_names(self):
        case_names = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[2]'):
            t = html.tostring(e, method='text', encoding='unicode')
# We strip "Errata: mm/dd/yyyy" from the case names of errata docs.
            if "Errata" in t:
                case_names.append(t.strip()[:-18])
# We strip "Public version posted on mm/dd/yyyy" from some case names.
            elif "posted on" in t:
                case_names.append(t.strip()[:-34])
            else:
                case_names.append(t.strip())
        return case_names

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[2]'):
            s = html.tostring(e, method='text', encoding='unicode').lower().strip()
            if "errata" in s:
                statuses.append('Errata')
            else:
                statuses.append('Published')
        return statuses

# This does not capture the release dates for the errata documents.
# The errata release date is listed in column 2. This will use the original
# release date instead.
    def _get_case_dates(self):
        case_dates = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[3]'):
            if html.tostring (e, method='text', encoding='unicode').strip() == "08/05/20008":
                date_string = "08/05/2008"
                case_dates.append(date.fromtimestamp(time.mktime(time.strptime(date_string, '%m/%d/%Y'))))
            else:
                date_string = html.tostring (e, method='text', encoding='unicode').strip()
                case_dates.append(date.fromtimestamp(time.mktime(time.strptime(date_string, '%m/%d/%Y'))))
        return case_dates

# Because there can be multiple docket numbers we have to replace some newlines.
    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[4]'):
            s = html.tostring (e, method='text', encoding='unicode').strip()
            docket_numbers.append(s.replace('\r\n', ' &'))
        return docket_numbers

# Comment out the following section if backscraping years 1999-2005.
# Because there are sometimes multiple judges we have to strip some whitespace.
    def _get_judges(self):
        judges = []
        for e in self.html.xpath('//table[3]/tr[position() > 1]/td[5]'):
            s = html.tostring (e, method='text', encoding='unicode')
            judges.append(s.strip())
        return judges



# Scraper for Minnesota Supreme Court
#CourtID: minn
#Court Short Name: MN
#Author: Andrei Chelaru
#Reviewer: mlr
#Date: 2014-07-03
#Contact:  Liz Reppe (Liz.Reppe@courts.state.mn.us), Jay Achenbach (jay.achenbach@state.mn.us)
import re
from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        # Get 500 most recent results for the year.  This number could be reduced after the
        # initial run of this fixed code.  We need ot start with a big number though to
        # capture the cases we've missed over the past few weeks.
        self.url = 'http://mn.gov/law-library-stat/search/opinions/?v:state=root%7Croot-0-500&query=date:{year}'.format(
            year=date.today().year
        )
        self.court_filters = ['/supct/']
        self.cases = []

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_case_data_from_html(html)
        return html

    def _extract_case_data_from_html(self, html):
        """Build list of data dictionaries, one dictionary per case.

        Sometimes the XML is malformed, usually because of a missing docket number,
        which throws the traditional data list matching off.  Its easier and cleaner
        to extract all the data at once, and simply skip over records that do not
        present a docket number.
        """
        for document in html.xpath('//document'):
            docket = document.xpath('content[@name="docket"]/text()')
            if docket:
                docket = docket[0]
                title = document.xpath('content[@name="dc.title"]/text()')[0]
                name = self._parse_name_from_title(title)
                url = document.xpath('@url')[0]
                if any(s in url for s in self.court_filters):
                    # Only append cases that are in the right jurisdiction.
                    self.cases.append({
                        'name': name,
                        'url': self._file_path_to_url(document.xpath('@url')[0]),
                        'date': convert_date_string(document.xpath('content[@name="date"]/text()')[0]),
                        'status': self._parse_status_from_title(title),
                        'docket': docket,
                    })

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _file_path_to_url(self, path):
        return path.replace('file:///web/prod/static/lawlib/live', 'http://mn.gov/law-library-stat')

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_precedential_statuses(self):
        return [case['status'] for case in self.cases]

    def _parse_name_from_title(self, title):
        name_regex = re.compile(r'(?P<name>.+)\s(?:A(?:DM)?\d\d-\d{1,4})', re.I)
        return name_regex.search(title).group('name')

    def _parse_status_from_title(self, title):
        return 'Unpublished' if 'unpublished' in title.lower() else 'Published'

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]



#  Scraper for Kansas Appeals Court
# CourtID: kanctapp
# Court Short Name: kanctapp
# Author: Andrei Chelaru
# Reviewer: mlr
# Date created: 25 July 2014


from juriscraper.opinions.united_states.state import kan


class Site(kan.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_index = 2



# Scraper for the United States Court of International Trade
# CourtID: cit
# Court Short Name: Ct. Int'l Trade
# Neutral Citation Format: Ct. Int'l Trade No. 12-1
from juriscraper.OpinionSite import OpinionSite
import re
import time
from datetime import date
from lxml import html

class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
# This is a special backscraper to deal with problems on the 2001 page.
        self.url = 'http://www.cit.uscourts.gov/SlipOpinions/SlipOps-2001.html'
        self.court_id = self.__module__

    def _get_download_urls(self):
        return [t for t in self.html.xpath('//table[3]//tr[position() > 1]/td//font//a/@href')]

    def _get_neutral_citations(self):
        neutral_citations = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[1]//font//a'):
            s = html.tostring(e, method='text', encoding='unicode').strip()
            neutral_citations.append(s)
        return neutral_citations

    def _get_case_names(self):
        case_names = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[2]/*'):
            s = html.tostring (e, method='text', encoding='unicode').strip()
# We strip "erratum: mm/dd/yyyy" from the case names of errata docs.
            if "errat" in s:
                case_names.append(s.strip()[:-19])
            else:
                case_names.append(s.strip())
        return case_names

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_case_dates(self):
        case_dates = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[3]//font'):
            s = html.tostring (e, method='text', encoding='unicode').strip()
            if s == '08//02/2001':
                case_dates.append(date.fromtimestamp(time.mktime(time.strptime('08/02/2001', '%m/%d/%Y'))))
            else:
                case_dates.append(date.fromtimestamp(time.mktime(time.strptime(s.strip(), '%m/%d/%Y'))))
        return case_dates

# Because there can be multiple docket numbers we have to replace some newlines.
    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath('//table[3]//tr[position() > 1]/td[4]//font'):
            s = html.tostring (e, method='text', encoding='unicode').strip()
            docket_numbers.append(s.replace('\r\n', ' &'))
        return docket_numbers



"""
Scraper for Indiana Supreme Court archive before 6/6/2005
CourtID: ind
Court Short Name: Ind.
Auth: Mike Lissner <mike@freelawproject.org>
Reviewer:
History:
    2014-09-03: Renamed to ind_2005.py by janderse
"""
from juriscraper.OpinionSite import OpinionSite

import time
from datetime import date
from lxml import html


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.in.gov/judiciary/opinions/previous/archsup.html'
        self.court_id = self.__module__

    def _get_case_names(self):
        return [e for e in self.html.xpath("//table/tr/td[1]/font/a/text()")]

    def _get_case_dates(self):
        dates = []
        for link_string in self.html.xpath('//table/tr/td[4]/font/a/@href'):
            date_string = link_string.split('/')[-1].split('.')[0][:-2]
            if date_string == '12319701-fs':
                # special case (there's always *one*)
                dates.append(date.fromtimestamp(
                    time.mktime(time.strptime('123197', '%m%d%y'))))
            elif len(date_string) == 4:
                dates.append(date.fromtimestamp(
                    time.mktime(time.strptime(date_string + '98', '%m%d%y'))))
            elif len(date_string) == 5:
                dates.append(date.fromtimestamp(
                    time.mktime(time.strptime('0' + date_string, '%m%d%y'))))
            elif len(date_string) == 6:
                dates.append(date.fromtimestamp(
                    time.mktime(time.strptime(date_string, '%m%d%y'))))
            else:
                print(date_string)
        return dates

    def _get_download_urls(self):
        return [s for s in self.html.xpath("//table/tr/td[4]/font/a/@href")]

    def _get_docket_numbers(self):
        return [html.tostring(e, method='text')
                    for e in self.html.xpath("//table/tr/td[2]")]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



# Author: Michael Lissner
# Date created: 2013-05-23

from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        year = date.today().year
        self.url = 'http://www.courts.state.hi.us/opinions_and_orders/opinions?yr=%s' % year
        self.back_scrape_iterable = range(2010, year + 1)
        self.court_id = self.__module__
        self.target_id = 'S.Ct'
        self.cases = []

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self.extract_cases_from_table(html)
        return html

    def extract_cases_from_table(self, html):
        path_rows = '//div[@id="table-content"]/table/tbody/tr'
        path_target = './td[2]/div/span/text()'
        for row in html.xpath(path_rows):
            court = row.xpath(path_target)
            if court and court[0].strip() == self.target_id:
                case_text = row.xpath('./td[4]/p/text() | ./td[4]/p/span/text()')[0]
                court_text = row.xpath('./td[5]/text()')
                self.cases.append({
                    'url': row.xpath('./td[3]/a/@href')[0],
                    'date': row.xpath('./td[1]/text()')[0],
                    'name': case_text.split(' (')[0],
                    'docket': row.xpath('./td[3]/a/text()')[0],
                    'court': court_text[0] if court_text else False,
                })

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case['date']) for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_lower_courts(self):
        return [case['court'] for case in self.cases]

    def _download_backwards(self, year):
        if int(year) < 2010:
            raise Exception('the haw portal does not include pre-2010 cases')

        self.url = 'http://www.courts.state.hi.us/opinions_and_orders/opinions?yr=%s' % year
        self.html = self._download()

        # Loop over all pages
        pages = self.html.xpath('//a[@class="page-numbers"]/text()')
        if pages:
            last_page = int(pages[-1])
            for page in range(2, last_page + 1):
                self.url = 'http://www.courts.state.hi.us/opinions_and_orders/opinions/page/%d?yr=%s' % (page, year)
                self.html = self._download()



# Scraper for Supreme Court of Oklahoma
# CourtID: okla
# Court Short Name: OK
# Court Contact: webmaster@oscn.net
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-05

import re
from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.elements = []
        self.base_path = False
        self.year = date.today().year
        self.regex = r'([^,]+), (\d{2}.\d{2}.\d{4}), (.*)'
        self.url = 'http://www.oscn.net/applications/oscn/Index.asp?ftdb=STOKCSSC&year=%d&level=1' % self.year

    def set_base_path(self):
        # All test files should be edited to use year value of 2018
        year = 2018 if self.test_mode_enabled() else self.year
        self.base_path = "//a[contains(./text(), '%d')]" % year

    def _download(self, request_dict={}):
        self.set_base_path()
        html = super(Site, self)._download(request_dict)
        self.elements = html.xpath(self.base_path)
        return html

    def _get_download_urls(self):
        path = "%s/@href" % self.base_path
        return self.html.xpath(path)

    def _get_case_dates(self):
        return [self._return_substring(e, 2) for e in self.elements]

    def _get_case_names(self):
        return [self._return_substring(e, 3) for e in self.elements]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_neutral_citations(self):
        return [self._return_substring(e, 1) for e in self.elements]

    def _return_substring(self, element, group):
        text = element.text_content()
        substring = re.search(self.regex, text).group(group)
        return substring if group != 2 else convert_date_string(substring)



from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    """This cour's website is implemented by Decisia, a software
    that has built-in anti-bot detection. If you make too many
    requests in rapid succession, the site will render a captcha
    page instead of rendering the real page content. Since the
    docket numbers for unpublished opinions are not published on
    the main results page (but on sub-info pages), we may (likely)
    need to issues multiple sub-requests for each crawl to retrieve
    the docket info for unpublished opinions.

    Given the above, we need to track and limit the number of
    sub-requests we make in order to prevent triggering the
    anti-bot detection, which would trigger the captcha and
    crash our scraping.

    If we ever wanted to implement a back scraper for this court,
    we'd likely need to create a separate class to handle it. In order
    to do this, we'd need to use the 'Advanced Search' mechanism to
    find all results. The results page will render 25 results. But,
    in a browser, when you scroll down, more results populate dynamically
    on the page via javascript.  So, we'd need to hit the page via a
    phantomjs webdriver and keep loading results. To make things more
    complicated, we'd also need to batch our (sub) requests in order
    to prevent triggering the anti-bot detection. I've testing sleeping
    for 1 minute every 40 requests, and that seems to work. Consequently,
    the back scraper would take a very very long time to run, but if we
    needed to do it, it can probably be done with the above method.
    """

    # fetch no more than 20 docket numbers for non-published opinions (see above)
    SUB_REQUEST_LIMIT = 20

    # The url query parameter below is required to actually render the data.
    # For whatever reason, the main page (url without IFRAME_QUERY) loads itself
    # with IFRAME_QUERY into an iframe to render the results. Don't ask why.
    IFRAME = 'iframe=true'

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://nmonesource.com/nmos/nmsc/en/nav_date.do?' + self.IFRAME
        self.cases = []

    def _download(self, request_dict={}):
        self.request_dict = request_dict
        html = super(Site, self)._download(self.request_dict)
        self.extract_cases(html)
        return html

    def extract_cases(self, html):
        sub_request_count = 0
        sub_page_docket_path = '//tr[contains(./td[@class="label"]/text(), "Docket")]/td[@class="metadata"]'
        for record in html.xpath('//div[@class="documentList"]//li'):
            url = record.xpath('.//div[@class="documents"]/a/@href')[0]
            info = record.xpath('.//div[@class="subinfo"]')[0]
            date_string = info.xpath('.//span[@class="publicationDate"]')[0].text_content()
            court_status_string = info.xpath('.//div[@class="subMetadata"]')[0].text_content()
            status = 'Unpublished' if 'unreported' in court_status_string.lower() else 'Published'
            anchor = info.xpath('.//span[@class="title"]/a')[0]
            name = anchor.text_content()
            citation = info.xpath('.//span[@class="citation"]')
            if citation:
                # the docket is displayed on the main results page (published opinion)
                docket = citation[0].text_content()
            else:
                if sub_request_count >= self.SUB_REQUEST_LIMIT:
                    # in order to avoid anti-bot detection, only don't fetch
                    # too many docket numbers for un-published opinions
                    continue
                if self.test_mode_enabled():
                    # avoid hitting network during test
                    docket = 'test-docket-placeholder'
                else:
                    # find the docket from the sub page (unpublished opinion)
                    self.url = '%s?%s' % (anchor.attrib['href'], self.IFRAME)
                    logger.info('%s: searching for docket on sub page %s' % (self.court_id, self.url))
                    sub_page = super(Site, self)._download(self.request_dict)
                    sub_request_count += 1
                    cell = sub_page.xpath(sub_page_docket_path)
                    if not cell:
                        # if no docket found on sub page, move on to the next opinion
                        logger.info('no docket found on sub page, skipping opinion "%s"' % name)
                        continue
                    docket = cell[0].text_content().strip()
                    logger.info('found docket "%s" on sub page for opinion "%s"' % (docket, name))
            self.cases.append({
                'name': name,
                'url': url,
                'docket': docket,
                'date': convert_date_string(date_string),
                'status': status,
            })

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_precedential_statuses(self):
        return [case['status'] for case in self.cases]

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]



"""
Scraper for Alaska Court of Appeals
ID: alaskactapp
Court Short Name: Alaska Court of Appeals
"""

from . import alaska


class Site(alaska.Site):
    url_court_id = 'ap'

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__



# Author: Michael Lissner
# Date created: 2013-06-06

import re
import requests

from datetime import date
from datetime import datetime

from juriscraper.opinions.united_states.state import nd
from juriscraper.DeferringList import DeferringList


class Site(nd.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        today = date.today()
        self.url = 'http://www.ndcourts.gov/opinions/month/%s.htm' % (today.strftime("%b%Y"))

    def _get_download_urls(self):
        """We use a fetcher and a DeferringList object and a HEAD request
        to test whether the wpd exists for a case"""
        def fetcher(html_link):
            if self.test_mode_enabled():
                return html_link    # Can't fetch remote during tests
            case_number = re.search(r'(\d+)', html_link).group(0)
            wpd_link = 'http://www.ndcourts.gov/wp/%s.wpd' % case_number
            r = requests.head(wpd_link,
                              allow_redirects=False,
                              headers={'User-Agent': 'Juriscraper'})
            if r.status_code == 200:
                return wpd_link
            else:
                return html_link

        if self.crawl_date >= date(1998, 10, 1):
            path = '//a/@href[contains(., "/court/opinions/")]'
            seed = list(self.html.xpath(path))
        else:
            path = '//ul//a[text()]/@href'
            seed = list(self.html.xpath(path))
        return DeferringList(seed=seed,
                             fetcher=fetcher)

    def _get_case_names(self):
        if self.crawl_date >= date(1998, 10, 1):
            path = '//a[contains(@href, "/court/opinions/")]/text()'
            return list(self.html.xpath(path))
        else:
            path = '//ul//a/text()'
            names = self.html.xpath(path)
            case_names = []
            if self.crawl_date < date(1996, 11, 1):
                # A bad time.
                for name in names:
                    name = name.rsplit('-')[0]
                    case_names.append(name)
                return case_names
            else:
                return list(names)

    def _get_case_dates(self):
        # A tricky one. We get the case dates, but each can have different number of cases below it, so we have to
        # count them.
        case_dates = []
        if self.crawl_date >= date(1998, 10, 1):
            test_path = '//body/a'
            if len(self.html.xpath(test_path)) == 0:
                # It's a month with no cases (like Jan, 2009)
                return []
            path = '//body/a|//body/font'
            for e in self.html.xpath(path):
                if e.tag == 'font':
                    date_str = e.text
                    dt = datetime.strptime(date_str, '%B %d, %Y').date()
                elif e.tag == 'a':
                    try:
                        case_dates.append(dt)
                    except NameError:
                        # When we don't yet have the date
                        continue
        else:
            path = '//h4|//li'
            for e in self.html.xpath(path):
                if e.tag == 'h4':
                    # We make dates on h4's because there's one h4 per date.
                    date_str = e.text.strip()
                    dt = datetime.strptime(date_str, '%B %d, %Y').date()
                elif e.tag == 'li':
                    try:
                        # We append on li's, because there's one li per case.
                        case_dates.append(dt)
                    except NameError:
                        # When we don't yet have the date
                        continue
        return case_dates

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        if self.crawl_date >= date(1998, 10, 1):
            path = '//a/@href[contains(., "/court/opinions/")]'
        else:
            path = '//ul//a[text()]/@href'
        docket_numbers = []
        for html_link in self.html.xpath(path):
            try:
                docket_numbers.append(re.search(r'(\d+)', html_link).group(0))
            except AttributeError:
                continue
        return docket_numbers

    def _get_neutral_citations(self):
        if self.crawl_date < date(1997, 2, 1):
            # Old format, but no neutral cites, thus short circuit the function.
            return None
        elif self.crawl_date < date(1998, 10, 1):
            # Old format with: 1997 ND 30 - Civil No. 960157 or 1997 ND 30
            path = '//li/text()'
        elif self.crawl_date >= date(1998, 10, 1):
            # New format with: 1997 ND 30
            path = '//body/text()'
        neutral_cites = []
        for t in self.html.xpath(path):
            try:
                neutral_cites.append(re.search(r'^.{0,5}(\d{4} ND (?:App )?\d{1,4})', t, re.MULTILINE).group(1))
            except AttributeError:
                continue
        return neutral_cites

    def _post_parse(self):
        # Remove any information that applies to non-appellate cases.
        if self.neutral_citations:
            delete_items = []
            for i in range(0, len(self.neutral_citations)):
                if 'App' in self.neutral_citations[i]:
                    delete_items.append(i)

            for i in sorted(delete_items, reverse=True):
                del self.download_urls[i]
                del self.case_names[i]
                del self.case_dates[i]
                del self.precedential_statuses[i]
                del self.docket_numbers[i]
                del self.neutral_citations[i]
        else:
            # When there aren't any neutral cites that means they're all supreme court cases.
            pass

    def _download_backwards(self, d):
        self.crawl_date = d
        self.url = 'http://www.ndcourts.gov/opinions/month/%s.htm' % (d.strftime("%b%Y"))
        self.html = self._download()



# -*- coding: utf-8 -*-
"""
Python 2.x Regular Express patterns for string_utils.py
"""
import re

BIG = ('3D|AFL|AKA|A/K/A|BMG|CBS|CDC|CDT|CEO|CIO|CNMI|D/B/A|DOJ|DVA|EFF|FCC|'
           'FTC|HSBC|IBM|II|III|IV|JJ|LLC|LLP|MCI|MJL|MSPB|ND|NLRB|PTO|SD|UPS|RSS|SEC|UMG|US|USA|USC|'
           'USPS|WTO')
SMALL = u'a|an|and|as|at|but|by|en|for|if|in|is|of|on|or|the|to|v\.?|via|vs\.?'
NUMS = u'0123456789'
PUNCT = ur"""!"#$¢%&'‘()*+,\-./:;?@[\\\]_—`{|}~"""
WEIRD_CHARS = ur'¼½¾§ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÑÒÓÔÕÖØÙÚÛÜßàáâãäåæçèéêëìíîïñòóôœõöøùúûüÿ'
BIG_WORDS = re.compile(ur'^(%s)[%s]?$' % (BIG, PUNCT), re.I | re.U)
SMALL_WORDS = re.compile(r'^(%s)$' % SMALL, re.I)
SMALL_WORD_INLINE = re.compile(r'(^|\s)(%s)(\s|$)' % SMALL, re.I | re.U)
INLINE_PERIOD = re.compile(r'[a-z][.][a-z]', re.I)
INLINE_SLASH = re.compile(r'[a-z][/][a-z]', re.I)
INLINE_AMPERSAND = re.compile(r'([a-z][&][a-z])(.*)', re.I)
UC_ELSEWHERE = re.compile(ur'[%s]*?[a-zA-Z]+[A-Z]+?' % PUNCT, re.U)
CAPFIRST = re.compile(ur"^[%s]*?([A-Za-z])" % PUNCT)
SMALL_FIRST = re.compile(ur'^([%s]*)(%s)\b' % (PUNCT, SMALL), re.I | re.U)
SMALL_LAST = re.compile(r'\b(%s)[%s]?$' % (SMALL, PUNCT), re.I | re.U)
SUBPHRASE = re.compile(r'([:;?!][ ])(%s)' % SMALL)
APOS_SECOND = re.compile(r"^[dol]{1}['‘]{1}[a-z]+$", re.I)
ALL_CAPS = re.compile(ur'^[A-Z\s%s%s%s]+$' % (PUNCT, WEIRD_CHARS, NUMS))
UC_INITIALS = re.compile(r"^(?:[A-Z]{1}\.{1}|[A-Z]{1}\.{1}[A-Z]{1})+,?$")
MAC_MC = re.compile(r'^([Mm]a?c)(\w+.*)')


# Back Scraper for New York Appellate Divisions 4th Dept.
# CourtID: nyappdiv_4th
# Court Short Name: NY
# Author: Andrei Chelaru
# Reviewer:
# Date: 2015-10-30

from juriscraper.opinions.united_states_backscrapers.state import ny


class Site(ny.Site):

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court = 'App+Div,+4th+Dept'
        self.interval = 30



"""Scraper for Kentucky Supreme Court
CourtID: ky
Court Short Name: Ky.
Contact: https://courts.ky.gov/aoc/technologyservices/Pages/ContactTS.aspx

History:
    2014-08-07: Created by mlr.
    2014-12-15: Updated to fetch 100 results instead of 30. For reasons unknown
                this returns more recent results, while a smaller value leaves
                out some of the most recent items.
    2016-03-16: Restructured by arderyp to pre-process all data at donwload time
                to work around resources whose case names cannot be fetched to
                to access restrictions.
Notes:
    This scraper is unique. Kentucky does not provide case names in the primary
    search portal's result page, making them almost useless. They have a secondary
    search portal that allows a lookup by case year and number, which *does* provide
    the case name (and lots of other information). Note that it only provides
    this information for supreme court cases, so extending this to do kyctapp
    won't be possible.

    Primary Search Portal:          http://apps.courts.ky.gov/supreme/sc_opinions.shtm
    Primary Search Portal POST:     http://162.114.92.72/dtsearch.asp
    Secondary Search Portal:        https://appellate.kycourts.net/SC/SCDockets/
    Secondary Search Portal GET:    https://appellate.kycourts.net/SC/SCDockets/CaseDetails.aspx?cn={yyyySC######}

    Our two step process is as follows:
      1. Get the pdf url, case date, and docket number from the Primary Search Portal
      2. If the above data is valid, use publication year and docket number
         to fetch case name from Secondary Search Portal

    Unlike other scrapers, all of the metadata above is gathered at download time,
    saved on the local object, and merely echoed back in by the standard getters.
    This design allows us to avoid issues with cases that have restricted access
    in the Secondary Search Portal (example: year=2015, docket=000574).  These
    restricted resources will be skipped.

    Also fun, they use IP addresses instead of DNS and hide them behind HTML
    frames hosted by real domains.

    I tried calling to get more information, and though they've heard of us (a
    first!), they didn't want to help, and seemed downright aggressive in their
    opposition. Curious. Anyway, don't bother calling  again.

    You can contact support@dtsearch.com with questions about the search
    interface. Best of luck.
"""

import re

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import titlecase
from juriscraper.lib.string_utils import convert_date_string

## WARNING: THIS SCRAPER IS FAILING:
## This scraper is succeeding in development, but
## is failing in production.  We are not exactly
## sure why, and suspect that the hosting court
## site may be blocking our production IP and/or
## throttling/manipulating requests from production.


class Site(OpinionSite):
    CASE_NAMES = []
    CASE_DATES = []
    DOWNLOAD_URLS = []
    DOCKET_NUMBERS = []

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        # Sometimes they change the ip/endpoint below, in which
        # case the scraper will throw a 404 error. just go to
        # http://apps.courts.ky.gov/supreme/sc_opinions.shtm
        # and inspect the form html and grab the new "action" url
        self.url = 'http://162.114.92.72/dtSearch/dtisapi6.dll'
        # Sometimes they also change up the key/value pairs below.
        # What you want to do is go to the url in the comment above,
        # open your browser inspector's network tab, then search on
        # the page's form for "xfirstword", 100 results, sorted by
        # date, click search, and then inspect the Params for the
        # POST request that was just triggered. This site sucks.
        self.parameters = {
            'index': "*{aa61be39717dafae0e114c24b74f68db}+Supreme+Court+Opinions+(1996+)",
            'request': "xfirstword",
            'fuzziness': "0",
            'MaxFiles': "100",
            'autoStopLimit': "0",
            'sort': "Date",
            'cmd': "search",
            'SearchForm': "%%SearchForm%%",
            'dtsPdfWh': "*",
            'OrigSearchForm': "/dtsearch_form.html",
            'pageSize': "1000",
            'fileConditions': "",
            'booleanConditions': "",
        }
        self.method = 'POST'
        self.docket_number_regex = re.compile(r'(?P<year>\d{4})-(?P<court>[SC]{2})-(?P<number>\d+)')
        self.hrefs_contain = 'Opinions'

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._build_data_lists_from_html(html)
        return html

    def _build_data_lists_from_html(self, html):
        # Search second column cells for valid opinions
        data_cell_path = '//table/tr/td[2]/font'
        for cell in html.xpath(data_cell_path):
            # Cell must contain a link
            if cell.xpath('a'):
                link_href = cell.xpath('a/@href')[0].strip()
                link_text = cell.xpath('a/text()')[0].strip()
                cell_text = cell.xpath('text()')
                date = self._parse_date_from_cell_text(cell_text)
                # Cell must contain a parse-able date
                if date:
                    docket_match = self.docket_number_regex.search(link_text)
                    # Cell must contain a docket number conforming to expected format
                    if docket_match:
                        if self.test_mode_enabled():
                            # Don't fetch names when running tests
                            name = 'No case names fetched during tests.'
                        else:
                            # Fetch case name from external portal search (see doc string at top for details)
                            case_number = '%s%s%s' % (docket_match.group('year'), docket_match.group('court'), docket_match.group('number'))
                            name = self._fetch_case_name(case_number)
                        if name:
                            docket_number = '%s %s %s' % (
                                docket_match.group('year'),
                                docket_match.group('court'),
                                docket_match.group('number'),
                            )
                            self.CASE_NAMES.append(name)
                            self.CASE_DATES.append(date)
                            self.DOCKET_NUMBERS.append(docket_number)
                            self.DOWNLOAD_URLS.append(link_href)

    def _parse_date_from_cell_text(self, cell_text):
        date = False
        for text in cell_text:
            try:
                date = convert_date_string(text.strip())
                break
            except ValueError:
                pass
        return date

    def _fetch_case_name(self, case_number):
        """Fetch case name for a given docket number + publication year pair.

        Some resources show 'Public Access Restricted' messages and do not
        provide parseable case name information.  These will be skipped by
        our system by returning False below.  The only other approach would
        be to parse the case name from the raw PDF text itself.
        """

        # If case_number is not expected 12 characters, skip it, since
        # we can't know how to fix the courts typo. They likely forgot
        # to '0' pad the beginning or the end of the 'number' suffix,
        # but we can't know for sure.
        if len(case_number) != 12:
            return False

        # Site has non-chained, bad certificate, need to
        # ignore ssl verification for now for scraper to work
        self.request['verify'] = False

        url = 'https://appellate.kycourts.net/SC/SCDockets/CaseDetails.aspx?cn=%s' % case_number
        html = self._get_html_tree_by_url(url)

        # Halt if there is a (dismissible) error/warning on the page
        path_error_warning = '//div[contains(@class, "alert-dismissible")]'
        if html.xpath(path_error_warning):
            raise InsanityException('Invalid sub-resource url (%s). Is case number (%s) invalid?' % (url, case_number))

        # Ensure that only two substrings are present
        path_party = '//td[@class="party"]/text()'
        parties = html.xpath(path_party)
        if len(parties) != 2:
            raise InsanityException('Unexpected party elements. Expected two substrings, got: %s' % ', '.join(parties))

        return titlecase(' v. '.join(parties))

    def _get_download_urls(self):
        return self.DOWNLOAD_URLS

    def _get_case_names(self):
        return self.CASE_NAMES

    def _get_docket_numbers(self):
        return self.DOCKET_NUMBERS

    def _get_case_dates(self):
        return self.CASE_DATES

    def _get_precedential_statuses(self):
        return ['Unknown'] * len(self.CASE_NAMES)



# Scraper for the United States Court of Appeals for the Seventh Circuit
# CourtID: ca7
# Court Short Name: 7th Cir.

import time
from six.moves.urllib.parse import urlencode
from datetime import date, timedelta
from dateutil.rrule import rrule, DAILY
from lxml import html

from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.a_while_ago = date.today() - timedelta(days=60)
        self.url = 'http://media.ca7.uscourts.gov/cgi-bin/rssExec.pl?Time=any&FromMonth={month}&FromDay={day}&FromYear={year}&' \
                   'ToMonth=&ToDay=&ToYear=&Author=any&AuthorName=&Case=any&CaseY1=&CaseY2=&CaseN1=&CaseN2=&CaseN3=&' \
                   'CaseN4=&Submit=Submit&RssJudgeName=Easterbrook&OpsOnly=no'.format(
            month=self.a_while_ago.month,
            day=self.a_while_ago.day,
            year=self.a_while_ago.year,
        )
        self.court_id = self.__module__

        self.interval = 30
        self.back_scrape_iterable = [i.date() for i in rrule(
            DAILY,
            interval=self.interval,  # Every interval days
            dtstart=date(1999, 10, 1),
            until=date(2015, 1, 1),
        )]

    def _get_case_names(self):
        case_names = []
        for e in self.html.xpath('//table//table/tr[position() >= 3]/td[2]'):
            s = html.tostring(e, method='text', encoding='unicode')
            case_names.append(s)
        return case_names

    def _get_download_urls(self):
        return [e for e in self.html.xpath('//table[2]/tr/td/table/tr[position() >=3]/td/a/@href')]

    def _get_case_dates(self):
        return [date.fromtimestamp(time.mktime(time.strptime(date_string.strip(), '%m/%d/%Y')))
                for date_string in self.html.xpath('//table//table/tr[position() >= 3]/td[4]/text()')]

    def _get_docket_numbers(self):
        return [docket_number for docket_number in
                self.html.xpath('//table//table/tr[position() >= 3]/td[1]/text()')]

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath('//table//table/tr[position() >= 3]/td[5]/a'):
            s = html.tostring(e, method='text', encoding='unicode')
            if 'Opinion' in s:
                statuses.append('Published')
            elif 'Nonprecedential' in s:
                statuses.append('Unpublished')
            else:
                statuses.append('Unknown')
        return statuses

    def _get_nature_of_suit(self):
        natures = []
        for e in self.html.xpath('//table//table/tr[position() >= 3]/td[3]'):
            natures.append(html.tostring(e, method='text', encoding='unicode'))
        return natures

    def _get_judges(self):
        judges = []
        for e in self.html.xpath('//table//table/tr[position() >= 3]/td[6]'):
            s = html.tostring(e, method='text', encoding='unicode')
            if s.lower().strip() == 'percuriam':
                s = "Per Curiam"
            judges.append(s)
        return judges

    def _download_backwards(self, d):
        to_date = d + timedelta(self.interval)
        params = urlencode({
            'Time': 'any',
            'FromMonth': d.month,
            'FromDay': d.day,
            'FromYear': d.year,
            'ToMonth': to_date.month,
            'ToDay': to_date.day,
            'ToYear': to_date.year,
            'Author': 'any',
            'AuthorName': '',
            'Case': 'any',
            'CaseY1': '',
            'CaseY2': '',
            'CaseN1': '',
            'CaseN2': '',
            'CaseN3': '',
            'CaseN4': '',
            'Submit': 'Submit',
            'RssJudgeName': 'Easterbrook',
            'OpsOnly': 'no'
        })
        self.base_url = 'http://media.ca7.uscourts.gov/cgi-bin/rssExec.pl'
        self.url = "{}?{}".format(self.base_url, params)

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



# Auth: mlr
# Date: 2013-05-28

from juriscraper.opinions.united_states.state import hawapp

from datetime import date
from lxml import html


class Site(hawapp.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__

        # If it's the beginning of January, we need to make sure that we aren't
        # missing any late-coming cases from the previous year.
        today = date.today()
        if today.day < 15 and today.month == 1:
            this_year = today.year
            last_year = this_year - 1
            self.url = self.url.replace(str(this_year), str(last_year))
        else:
            # This simply aborts the crawler.
            self.status = 200
            self.html = html.fromstring('<html></html>')



from datetime import timedelta, date
from dateutil.rrule import rrule, MONTHLY

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.start = date.today() - timedelta(days=60)
        self.end = date.today()
        self.url = 'http://www.courts.wa.gov/opinions/index.cfm?fa=opinions.processSearch'
        self.method = 'POST'

        self.courtLevel = 'S'
        self.pubStatus = 'All'
        self._set_parameters()
        self.base = "//tr[../tr/td/strong[contains(., 'File Date')]]"
        self.back_scrape_iterable = [i.date() for i in rrule(
            MONTHLY,
            dtstart=date(2014, 2, 1),
            until=date.today(),
        )]

    def _set_parameters(self):
        self.parameters = {
            'courtLevel': self.courtLevel,
            'pubStatus': self.pubStatus,
            'beginDate': self.start.strftime('%m/%d/%Y'),
            'endDate': self.end.strftime('%m/%d/%Y'),
            'SType': 'Phrase',
            'SValue': ''
        }

    def _get_case_names(self):
        path = "{base}/td[3]/text()".format(base=self.base)
        return list(self.html.xpath(path))

    def _get_docket_numbers(self):
        path = "{base}/td[2]/a/text()".format(base=self.base)
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = "{base}/td[1]/text()".format(base=self.base)
        return [convert_date_string(date.strip()) for date in self.html.xpath(path) if date.strip()]

    def _get_download_urls(self):
        path = "{base}/td[2]/a[2]/@href".format(base=self.base)
        return list(self.html.xpath(path))

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _download_backwards(self, d):
        self.start = d
        self.end = d + timedelta(days=32)
        self._set_parameters()
        self.html = self._download()






# Author: Michael Lissner
# Date created: 2013-06-21

from juriscraper.opinions.united_states.state import neb


class Site(neb.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'https://supremecourt.nebraska.gov/courts/court-appeals/opinions'



# Scraper for the United States Tax Court
# CourtID: tax
# Court Short Name: Tax Ct.
# Neutral Citation Format (Tax Court opinions): 138 T.C. No. 1 (2012)
# Neutral Citation Format (Memorandum opinions): T.C. Memo 2012-1
# Neutral Citation Format (Summary opinions: T.C. Summary Opinion 2012-1

import os
from datetime import date, datetime, timedelta

from dateutil.rrule import WEEKLY, rrule
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from juriscraper.AbstractSite import logger, phantomjs_executable_path
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.html_utils import fix_links_but_keep_anchors


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.uses_selenium = True
        self.url = 'https://www.ustaxcourt.gov/UstcInOp/OpinionSearch.aspx'
        self.base_path = '//tr[@class="ResultsOddRow" or ' \
                         '@class="ResultsEvenRow"]'
        self.case_date = date.today()
        self.backwards_days = 14
        self.back_scrape_iterable = [i.date() for i in rrule(
            WEEKLY,
            dtstart=date(1995, 9, 25),
            until=date(2018, 11, 13),
        )]

        self.court_id = self.__module__

    def _download(self, request_dict={}):
        """Uses Selenium because doing it with requests is a pain."""
        if self.test_mode_enabled():
            return super(Site, self)._download(request_dict=request_dict)

        driver = webdriver.PhantomJS(
            executable_path=phantomjs_executable_path,
            service_log_path=os.path.devnull,  # Disable ghostdriver.log
        )
        driver.implicitly_wait(30)
        logger.info("Now downloading case page at: %s" % self.url)
        driver.get(self.url)

        # Set the start and end dates
        start_date_id = 'ctl00_Content_dpDateSearch_dateInput'
        start_date_input = driver.find_element_by_id(start_date_id)
        start_date_input.send_keys((self.case_date - timedelta(
            days=self.backwards_days)).strftime('%-m/%-d/%Y'))

        end_date_id = 'ctl00_Content_dpDateSearchTo_dateInput'
        end_date_input = driver.find_element_by_id(end_date_id)
        end_date_input.send_keys(self.case_date.strftime('%-m/%-d/%Y'))
        # driver.save_screenshot('%s.png' % self.case_date)

        # Check ordering by case date (this orders by case date, *ascending*)
        ordering_id = 'Content_rdoCaseName_1'
        driver.find_element_by_id(ordering_id).click()

        # Submit
        driver.find_element_by_id('Content_btnSearch').click()

        # Do not proceed until the results show up.
        wait = WebDriverWait(driver, 10)
        wait.until(EC.presence_of_element_located(
            (By.ID, 'Content_ddlResultsPerPage'))
        )
        # driver.save_screenshot('with-results.png')

        text = self._clean_text(driver.page_source)
        driver.quit()
        html_tree = self._make_html_tree(text)
        html_tree.rewrite_links(fix_links_but_keep_anchors,
                                base_href=self.url)
        return html_tree

    def _get_download_urls(self):
        # URLs here take two forms. For precedential cases, they're simple, and
        # look like:
        #   https://www.ustaxcourt.gov/UstcInOp/OpinionViewer.aspx?ID=11815
        # But for non-precedential cases, they have an annoying JS thing, and
        # links look like:
        #   https://www.ustaxcourt.gov/UstcInOp/OpinionSearch.aspx#11813
        # Note the hash instead of the ID.
        #
        # This is annoying, but we just have to swap out the ending and it
        # should be fine.
        hrefs = []
        path = self.base_path + '//@href'
        for href in self.html.xpath(path):
            if '?ID' in href:
                hrefs.append(href)
            else:
                hrefs.append(href.replace('OpinionSearch.aspx#',
                                          'OpinionViewer.aspx?ID='))
        return hrefs

    def _get_case_names(self):
        case_names = []
        path = self.base_path + '//td[1]'
        for td in self.html.xpath(path):
            case_names.append(td.text_content().strip() + ' v. Commissioner')
        return case_names

    def _get_case_name_shorts(self):
        # The normal values are particularly bad, usually just returning
        # "Commissioner" for all cases. Just nuke these values.
        return [''] * len(self.case_names)

    def _get_precedential_statuses(self):
        statuses = []
        path = self.base_path + '//td[2]'
        for td in self.html.xpath(path):
            status = td.text_content().strip().lower()
            if "memorandum" in status:
                statuses.append('Published')
            elif "summary" in status:
                statuses.append('Unpublished')
            else:
                statuses.append('Unknown')
        return statuses

    def _get_case_dates(self):
        dates = []
        path = self.base_path + '//td[3]'
        for td in self.html.xpath(path):
            date_string = td.text_content().strip()
            dates.append(datetime.strptime(date_string, '%m/%d/%Y').date())
        return dates

    def _download_backwards(self, d):
        self.backwards_days = 7
        self.case_date = d
        logger.info("Running backscraper with date range: %s to %s",
                    self.case_date - timedelta(days=self.backwards_days),
                    self.case_date)
        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



from datetime import date
from datetime import datetime
from datetime import timedelta

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import clean_if_py3
from dateutil.rrule import rrule, DAILY


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.ca4.uscourts.gov/DataOpinions.aspx'
        self.court_id = self.__module__
        td = date.today()
        self.parameters = {
            'CASENUM': '',
            'FROMDATE': (td - timedelta(days=10)).strftime('%m-%d-%Y'),
            'TITLE': '',
            'TODATE': td.strftime('%m-%d-%Y')
        }
        self.method = 'POST'
        self.interval = 30
        self.back_scrape_iterable = [i.date() for i in rrule(
            DAILY,
            interval=self.interval,  # Every interval days
            dtstart=date(1996, 1, 1),
            until=date(2016, 1, 1),
        )]

    def _get_case_names(self):
        path = '//tr/td[4]/text()'
        names = []
        for s in self.html.xpath(path):
            s = clean_if_py3(s)
            if s.strip():
                names.append(s)
        logger.info(str(len(names)))
        return names

    def _get_download_urls(self):
        path = '//tr/td[1]/a/@href'
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = '//tr/td[3]/text()'
        case_dates = []
        for date_string in self.html.xpath(path):
            date_string = clean_if_py3(date_string).strip()
            if date_string:
                case_dates.append(datetime.strptime(date_string, '%m/%d/%Y').date())
        return case_dates

    def _get_docket_numbers(self):
        path = '//tr/td[2]//text()'
        docket_numbers = []
        for s in self.html.xpath(path):
            s = clean_if_py3(s).strip()
            if s:
                docket_numbers.append(s)
        return docket_numbers

    def _get_precedential_statuses(self):
        statuses = []
        # using download link, we can get the statuses
        for download_url in self.download_urls:
            file_name = download_url.split('/')[-1]
            if 'u' in file_name.lower():
                statuses.append('Unpublished')
            else:
                statuses.append('Published')
        return statuses

    def _download_backwards(self, d):

        self.parameters = {
            'CASENUM': '',
            'FROMDATE': d.strftime('%m-%d-%Y'),
            'TITLE': '',
            'TODATE': (d + timedelta(self.interval)).strftime('%m-%d-%Y')
        }

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



"""Scraper for Maryland Supreme Court Oral Argument Audio

This scraper has an interesting history. It was briefly running on the live
site, but we realized shortly after starting it that the scraper was
downloading video, not audio!

Seeing that we weren't ready for video, we disabled this scraper and deleted
any traces of it on the server.

One interesting lesson though was that the OA system didn't crumble or budge
when this was running. The video converted to mp3 just fine (each item took a
few hours) and we began hosting it like nothing was different. Go figure.

Your humble editor,

Mike


CourtID: md
Court Short Name: Md.
Author: Brian W. Carver
Reviewer: mlr
History:
  2014-10-17: Created by Brian Carver
  2014-10-20: Some tweaks by mlr
"""

from datetime import datetime

from juriscraper.OralArgumentSite import OralArgumentSite


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.courts.state.md.us/coappeals/webcasts/webcastarchive.html'

    def _get_download_urls(self):
        path = "//table//tr/td[2]/strong/a/@href[not(contains(., 'baradmissions'))]"
        # This works because they usually link on docket number, but there's
        # one page where the audio must have been long and they link instead
        # on 'Part 1' and 'Part 2'. That sort of irregularity is not handled.
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        dates = []
        path = "//table/tr[not(contains(.//@href, 'baradmission'))]/td[1]/text()"
        for s in self.html.xpath(path):
            for date_format in ('%m-%d-%Y', '%m-%d-%y'):
                try:
                    d = datetime.strptime(s, date_format).date()
                except ValueError:
                    continue
                dates.append(d)
        return dates

    def _get_case_names(self):
        # To avoid the "Title" rows we find the siblings of the rows with links.
        path = '//table//tr/td[2]/strong/a/../../following-sibling::td/strong/text()'
        cases = []
        for case in self.html.xpath(path):
            if case.strip():
                cases.append(case)
        return cases

    def _get_docket_numbers(self):
        # This works except for on the page where one of the recordings was
        # split into Part 1 and Part 2.
        path = "//table//tr/td[2]/strong/a/text()[not(contains(., 'Admissions'))]"
        return list(self.html.xpath(path))



"""Scraper for Tenth Circuit of Appeals
CourtID: ca10
Author: mlr
"""

from juriscraper.OralArgumentSite import OralArgumentSite
from juriscraper.lib.string_utils import convert_date_string
from juriscraper.lib.html_utils import get_table_column_text, get_table_column_links


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.ca10.uscourts.gov/clerk/oral-argument-recordings'

    def _get_download_urls(self):
        return get_table_column_links(self.html, 5)

    def _get_case_names(self):
        return get_table_column_text(self.html, 2)

    def _get_docket_numbers(self):
        dockets = get_table_column_text(self.html, 1)
        # Handle case where "no results" message is put in table cell, see: ca10_example_2.html
        return [] if len(dockets) == 1 and 'no results' in dockets[0].lower() else dockets

    def _get_case_dates(self):
        return [convert_date_string(ds) for ds in get_table_column_text(self.html, 3)]



"""Scraper for Connecticut Appellate Court
CourtID: connctapp
Court Short Name: Connappct.
Author: Asadullah Baig<asadullahbeg@outlook.com>
Date created: 2014-07-11
"""

from datetime import date
from juriscraper.opinions.united_states.state import conn


class Site(conn.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.crawl_date = date.today()
        self.url = 'http://www.jud.ct.gov/external/supapp/archiveAROap{year}.htm'.format(
            year=self.crawl_date.strftime("%y"))
        self.court_id = self.__module__



from itertools import chain, islice, tee
try:
    from itertools import izip
except ImportError:
    izip = zip


def previous_and_next(some_iterable):
    """Provide previous and next values while iterating a list.

    This is from: http://stackoverflow.com/a/1012089/64911

    This will allow you to lazily iterate a list such that as you iterate, you
    get a tuple containing the previous, current, and next value.
    """
    prevs, items, nexts = tee(some_iterable, 3)
    prevs = chain([None], prevs)
    nexts = chain(islice(nexts, 1, None), [None])
    return izip(prevs, items, nexts)



"""Scraper for United States District Court for the District of Columbia
CourtID: dcd
Court Short Name: D.D.C.
Author: V. David Zvenyach
Date created: 2014-02-27
Substantially Revised: Brian W. Carver, 2014-03-28
"""

from juriscraper.opinions.united_states_backscrapers.federal_district import dcd_2013

class Site(dcd_2013.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://ecf.dcd.uscourts.gov/cgi-bin/Opinions.pl?2012'



"""Scraper for the Rhode Island Supreme Court
CourtID: ri
Court Short Name: R.I.
Court Contact: helpdesk@courts.ri.gov, MFerris@courts.ri.gov (Ferris, Mirella), webmaster@courts.ri.gov
    https://www.courts.ri.gov/PDF/TelephoneDirectory.pdf
Author: Brian W. Carver
Date created: 2013-08-10
"""

import re
from datetime import datetime

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = 'http://www.courts.ri.gov/Courts/SupremeCourt/Pages/Opinions/Opinions'
        self.url = self.build_url()
        self.cases = []
        self.previous_date = None
        self.include_summary = True
        self.precedential_status = 'Published'

    def build_url(self):
        # This court hears things from mid-September to end of June. This
        # defines the "term" for that year, which triggers the website updates.
        today = datetime.today()
        if today >= datetime(today.year, 9, 15):
            year = today.year
        else:
            year = today.year - 1
        return '%s%d-%d.aspx' % (self.base_url, year, year + 1)

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self.extract_cases_from_html(html)
        return html

    def extract_cases_from_html(self, html):
        # case information spans over 3 rows, so must process 3 at a time:
        #   <tr> - contains case name, docket number, date and pdf link
        #   <tr> - contains case summary
        #   <tr> - contains a one-pixel gif spacer
        table = "//table[@id = 'onetidDoclibViewTbl0']/tr[position() > 1]"
        rows = list(html.xpath(table))
        row_triplets = list(zip(rows, rows[1:]))[::3]

        for tr1, tr2 in row_triplets:
            case = self.extract_case_from_rows(tr1, tr2)
            self.previous_date = case['date']
            self.cases.append(case)

    def extract_case_from_rows(self, row1, row2):
        docket = row1.xpath('./td/a/text()')[0]
        docket = ', '.join([d.strip() for d in docket.split(',')])
        url = row1.xpath("./td/a/@href")[0]
        text = row1.xpath("./td[1]/text()")[0]
        text_to_parse = [text]

        if self.include_summary:
            summary_lines = row2.xpath("./td/div/text()")
            summary = '\n'.join(summary_lines)
            joined_text = '\n'.join([text, summary_lines[0]])
            text_to_parse.append(joined_text)
        else:
            summary = False

        return {
            'url': url,
            'docket': docket,
            'date': self.parse_date_from_text(text_to_parse),
            'name': self.parse_name_from_text(text_to_parse),
            'summary': summary,
        }

    def parse_date_from_text(self, text_list):
        regex = '(.*?)(\((\w+\s+\d+\,\s+\d+)\))(.*?)'
        for text in text_list:
            date_match = re.match(regex, text)
            if date_match:
                return convert_date_string(date_match.group(3))

        # Fall back on previous case's date
        if self.previous_date:
            return self.previous_date

        raise InsanityException('Could not parse date from string, and no '
                                'previous date to fall back on: "%s"' %
                                text_list)

    @staticmethod
    def parse_name_from_text(text_list):
        regexes = [
            # Expected format
            '(.*?)(,?\sNos?\.)(.*?)',
            # Clerk typo, forgot "No."/"Nos." substring
            '(.*?)(,?\s\d+-\d+(,|\s))(.*?)',
            # Same as above, and there's an unconventional docket number
            # like 'SU-14-324' instead of '14-324'. See ri_p_example_4.html
            '(.*?)(,?\s(?:\w+-)?\d+-\d+(,|\s))(.*?)',
        ]

        for regex in regexes:
            for text in text_list:
                name_match = re.match(regex, text)
                if name_match:
                    return name_match.group(1)

        # "No."/"Nos." and docket missing, fall back on whatever's before first
        # semi-colon
        for text in text_list:
            if ';' in text:
                return text.split(';')[0]

        raise InsanityException('Could not parse name from string: "%s"' %
                                text_list)

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_precedential_statuses(self):
        return [self.precedential_status] * len(self.cases)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_summaries(self):
        return [case['summary'] for case in self.cases]



import argparse
import os
import pprint
import re
import sys
from datetime import date

import feedparser
from requests import Session

from .docket_report import DocketReport
from .utils import clean_pacer_object, get_pacer_case_id_from_nonce_url, \
    get_pacer_doc_id_from_doc1_url, get_pacer_seq_no_from_doc1_url
from ..lib.html_utils import html_unescape
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, harmonize

logger = make_default_logger()

"""
As of 2018-04-25, the jurisdictions below do not have functional RSS feeds. I 
reached out to all of these jurisdictions and heard back the following:
 
 - miwb: Sent email. Left a VM.
 - nceb: "The Judges in our District have given direction that they do not want 
   the RSS feed turned on in our court.  This has been discussed several times 
   and their decision remains the same."
    - Update via email: "At this time, the court declined to participate.  
      However, the court anticipates transitioning to the NextGen version of 
      CM/ECF in 2019 and will likely participate at that time.
 - nmib (Northern Mariana Islands, 17 hours ahead): Sent email.
 - alnd: Sent email.
 - caed: Sent email. 916-930-4000, left message with IT director, Richard.
 - flnd: Sent email.
 - gand: Left a message at the "Systems" department
 - gasd: Sent email.
 - hid: Sent email. 808-541-1304, Lian Abernathy, Chief Deputy 
 - ilsd: Sent email.
 - kyed: Need to call Sharon Drolijk (Operations Manager) at (859) 233-2503.
 - mdd: Spoke with "Attorney Adviser". He asked for a letter, which I sent. He 
   will take it up with the court, though they didn't have any policy that he 
   was aware of. It's going through their "court committee system for 
   consideration." Seems it wasn't done for performance reasons years ago. They
   will take it up again.
 - msnd: Sent email.
 - mtd: Julie Collins, Deputy Clerk in charge of Billings division, called back. 
   She's going to look into it.
 - ndd: Sent email. 701-530-2313, Clerk of Clerk, Rob. On the to do list. 
   Hopefully by end of the year (2018). Nobody ever asked before.
 - nvd: Sent email. Talked to Robert in "Quality Control". He shook me a bit 
   with how difficult he was. He insists I need to send a letter to the chief 
   judge, but that they're not considering this currently.
    - A different tack: Trying the operations manager, Lia at 775-686-5840. 
      She's friendly. Requested a letter for chief judge. Sent.
 - nywd: Sent email.
 - oked: Sent email.
 - oknd: Sent email.
 - pamd: Sent email.
 - scd: "We have forwarded your request on to management staff at the District 
   Court for further consideration." 
 - tnwd: Sent email.
 - txnd: Left me a long voicemail. They're working on it and have it in 
   committee. Expectation is that they may require an en banc meeting of the 
   judges to make a decision, but that the decision should come soon and be 
   positive.
 - vaed: "We are currently looking into this and will possibly have this 
   feature in the near future."

"""


class PacerRssFeed(DocketReport):
    # The entries are HTML entity-coded, and these matches are run AFTER
    # decoding. A typical entry is of the form:
    #   [short_description] (<a href="doc1_url">document_number</a>)
    # Or a literal example (line breaks added):
    #   [Scheduling Order] (<a href="https://ecf.mad.uscourts.gov/doc1/
    #   09518690740?caseid=186403&de_seq_num=98">39</a>)
    # We use three simple matches rather than a complex one with three groups.
    document_number_regex = re.compile(r'">(\d+)</a>')
    doc1_url_regex = re.compile(r'href="(.*/doc1/.*)"')
    short_desc_regex = re.compile(r'\[(.*?)\]')  # Matches 'foo': [ foo ] (

    PATH = 'cgi-bin/rss_outside.pl'

    CACHE_ATTRS = ['data']

    def __init__(self, court_id):
        super(PacerRssFeed, self).__init__(court_id)
        self._clear_caches()
        self._data = None
        self.session = Session()
        self.is_valid = True

        if self.court_id.endswith('b'):
            self.is_bankruptcy = True
        else:
            self.is_bankruptcy = False

    @property
    def url(self):
        if self.court_id == 'ilnb':
            return "https://tdi.ilnb.uscourts.gov/wwwroot/RSS/rss_outside.xml"
        else:
            return "https://ecf.%s.uscourts.gov/%s" % \
                (self.court_id, self.PATH)

    def query(self):
        """Query the RSS feed for a given court ID

        Note that we use requests here, and so we forgo some of the
        useful features that feedparser offers around the Etags and
        Last-Modified headers. This is fine for now because no PACER
        site seems to support these headers, but eventually we'll
        probably want to do better here. The reason we *don't* use
        feedparser already is because it presents a different set of
        APIs and Exceptions that we don't want to monkey with, at
        least, not yet, and especially not yet if PACER itself doesn't
        seem to care.

        For a good summary of this issue, see:
        https://github.com/freelawproject/juriscraper/issues/195#issuecomment-385848344
        """
        logger.info(u"Querying the RSS feed for %s" % self.court_id)
        timeout = (60, 300)
        self.response = self.session.get(self.url, timeout=timeout)

    def parse(self):
        self._clear_caches()
        self.response.raise_for_status()
        self._parse_text(self.response.text)

    def _parse_text(self, text):
        """Parse the feed and set self.feed

        :param text: The text of the RSS feed.
        :return None
        """
        self.feed = feedparser.parse(text)

    @property
    def data(self):
        """Override this to create a list of docket-like objects instead of the
         usual dict that is usually provided by the docket report.
        """
        if self._data is not None:
            return self._data

        data_list = []
        for entry in self.feed.entries:
            try:
                data = self.metadata(entry)
                data[u'parties'] = None
                data[u'docket_entries'] = self.docket_entries(entry)
            except AttributeError:
                # Happens when RSS items lack necessary attributes like a URL
                # or published date.
                pass
            else:
                if data[u'docket_entries'] and data['docket_number']:
                    data_list.append(data)

        self._data = data_list
        return data_list

    def metadata(self, entry):
        data = {
            u'court_id': self.court_id,
            u'pacer_case_id': get_pacer_case_id_from_nonce_url(entry.link),
            u'docket_number': self._get_docket_number(entry.title),
            u'case_name': self._get_case_name(entry.title),
            # Filing date is not available. Also the case for free opinions.
            u'date_filed': None,
            u'date_terminated': None,
            u'date_converted': None,
            u'date_discharged': None,
            u'assigned_to_str': '',
            u'referred_to_str': '',
            u'cause': '',
            u'nature_of_suit': '',
            u'jury_demand': '',
            u'demand': '',
            u'jurisdiction': '',
        }
        data = clean_pacer_object(data)
        return data

    @property
    def parties(self):
        raise NotImplementedError("No parties for RSS feeds.")

    def docket_entries(self, entry):
        """Parse the RSS item to get back a docket entry-like object.
        Although there is only one, return it as a list.
        """
        de = {
            u'date_filed': date(*entry.published_parsed[:3]),
            u'description': u'',
            u'document_number': self._get_value(self.document_number_regex,
                                                entry.summary) or None,
            u'short_description': html_unescape(
                self._get_value(self.short_desc_regex, entry.summary)),
        }

        doc1_url = self._get_value(self.doc1_url_regex, entry.summary)
        if doc1_url:
            de[u'pacer_doc_id'] = get_pacer_doc_id_from_doc1_url(doc1_url)
            de[u'pacer_seq_no'] = get_pacer_seq_no_from_doc1_url(doc1_url)
        else:
            # Some courts, in particular, NYED do not provide doc1 links and
            # instead provide show_case_doc links. Some docket entries don't
            # provide links. In either case, we can't provide pacer_doc_id.
            de[u'pacer_doc_id'] = u''
            de[u'pacer_seq_no'] = None

        return [de]

    def _get_docket_number(self, title_text):
        if self.is_bankruptcy:
            # Uses both b/c sometimes the bankr. cases have a dist-style docket
            # number.
            regexes = [self.docket_number_dist_regex,
                       self.docket_number_bankr_regex]
        else:
            regexes = [self.docket_number_dist_regex]
        for regex in regexes:
            match = regex.search(title_text)
            if match:
                return match.group(1)

    def _get_case_name(self, title_text):
        # 1:18-cv-04423 Chau v. Gorg &amp; Smith et al --> Chau v. Gorg & Smith
        try:
            case_name = title_text.split(' ', 1)[1]
        except IndexError:
            return u"Unknown Case Title"
        case_name = html_unescape(case_name)
        case_name = clean_string(harmonize(case_name))
        return case_name


def _main():
    # For help: python -m juriscraper.pacer.rss_feeds -h
    parser = argparse.ArgumentParser(
        prog="python -m %s.%s" %
        (__package__,
         os.path.splitext(os.path.basename(sys.argv[0]))[0]))
    parser.add_argument('-b', '--bankruptcy', action='store_true',
                        help='Use bankruptcy parser variant.')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('court_or_file', nargs='?', default='-',
                        help='''
A PACER court id or a local filename; defaults to stdin (-).
Any 3 or 4 character string is presumed to be a court;
sorry if that was your filename.''')

    args = parser.parse_args()

    arg_len = len(args.court_or_file)
    if 3 <= arg_len <= 4:
        feed = PacerRssFeed(args.court_or_file)
        print("Querying RSS feed at: %s" % feed.url)
        feed.query()
        print("Parsing RSS feed for %s" % feed.court_id)
        feed.parse()
    else:
        if not args.bankruptcy:
            feed = PacerRssFeed('fake_district_court_id')
        else:
            # final 'b' char is interpretted as bankruptcy
            feed = PacerRssFeed('fake_bankruptcy_court_id_b')
        if args.court_or_file == '-':
            print("Faking up RSS feed from stdin as %s" % feed.court_id)
            f = sys.stdin
        else:
            print("Reading RSS feed from %s as %s" %
                  (args.court_or_file, feed.court_id))
            f = open(args.court_or_file)
        feed._parse_text(f.read().decode('utf-8'))

    print("Got %s items" % len(feed.data))
    if args.verbose:
        print("Here they are:\n")
        pprint.pprint(feed.data, indent=2)


if __name__ == "__main__":
    _main()



"""Scraper for the United States Bankruptcy Appellate Panel for the Ninth Circuit
CourtID: bap9
Court Short Name: 9th Cir. BAP"""

import time
from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.ca9.uscourts.gov/bap/'
        self.court_id = self.__module__
        self.method = 'POST'
        self.parameters = {
            'c_mode': 'view',
            'c_page_size': '500',
        }

    def _get_case_names(self):
        path = '''//table[3]//tr/td[1]/a/text()'''
        return ["In re: " + titlecase(text) for text in self.html.xpath(path)]

    def _get_download_urls(self):
        path = '''//table[3]//tr/td[1]/a/@href'''
        return [e for e in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        statuses = []
        path = '''//table[3]//tr/td[2]//text()'''
        for txt in self.html.xpath(path):
            if 'Unpublished' in txt:
                statuses.append('Unpublished')
            elif 'Published' in txt:
                statuses.append('Published')
            else:
                statuses.append('Unknown')
        return statuses

    def _get_docket_numbers(self):
        path = '''//table[3]//tr/td[3]//text()'''
        return [docket_number for docket_number in self.html.xpath(path)]

    def _get_case_dates(self):
        path = '''//table[3]//tr/td[4]//text()'''
        return [date.fromtimestamp(time.mktime(time.strptime(date_string, '%m/%d/%Y')))
                    for date_string in self.html.xpath(path)]



# Scraper for Iowa Appeals Court
# CourtID: iowactapp
# Court Short Name: iowactapp

from . import iowa
from juriscraper.AbstractSite import logger


class Site(iowa.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.iowacourts.gov/iowa-courts/court-of-appeals/court-of-appeals-court-opinions/'

    def _download_backwards(self, _):
        """Overriding this method from parent iowa class because the
        site link/page structure for archived Court of Appeals opinions
        is different than that of Archived Supreme Court opinions
        """
        self.archive = True
        landing_page_html = self._download()
        path_filter_class = 'contains(./@class, "nav-link")'
        path_filter_text = 'contains(./text(), "Archived Court of Appeals Opinions")'
        path = '//a[%s][%s]/@href' % (path_filter_class, path_filter_text)
        for archive_page_url in landing_page_html.xpath(path):
            logger.info("Back scraping archive page: %s" % archive_page_url)
            archive_page_html = self._get_html_tree_by_url(archive_page_url)
            self.extract_archive_cases(archive_page_html)

    def extract_archive_cases(self, html):
        """Overriding this method from parent iowa class because the
        html structure for archived Court of Appeals opinion pages is
        different than that of Archived Supreme Court opinion pages
        """
        path_date = '//h2[contains(., "Summaries")]'
        for date_header in html.xpath(path_date):
            text = date_header.text_content()
            date_string = text.replace('Summaries', '').replace('Opinion', '')
            path_case_rows = './following::table[1]//tr[.//a]'
            for case_row in date_header.xpath(path_case_rows):
                self.cases.append({
                    'date': date_string,
                    'url': case_row.xpath('./td[1]//a/@href')[0],
                    'name': case_row.xpath('./td[2]')[0].text_content().strip(),
                    'docket': case_row.xpath('./td[1]')[0].text_content().strip(),
                })



# History:
#   2015-10-20: Created by Andrei Chelaru

from datetime import datetime
from lxml import html
from dateutil import parser
from juriscraper.OpinionSite import OpinionSite
from juriscraper.AbstractSite import logger


class Site(OpinionSite):
    def __init__(self):
        super(Site, self).__init__()
        self.year = 0
        self.court_id = self.__module__
        self.url = ''
        self.url_base = 'http://www.illinoiscourts.gov/Opinions/SupremeCourt/{year}/default.asp'
        self.back_scrape_iterable = range(1996, 2015)
        self.download_url_path = None
        self.case_name_path = None
        self.case_dates_path = None
        self.case_dates_path_day = None
        self.case_dates_path_month = None
        self.precedential_statuses_path = None
        self.docket_numbers_path = None
        self.neutral_citations_path = None

    def _get_download_urls(self):
        return list(self.html.xpath(self.download_url_path))

    def _get_case_names(self):

        case_names = [' '.join(i.strip() for i in e.xpath('.//text()')).strip() for e in self.html.xpath(self.case_name_path)]
        return case_names

    def _get_case_dates(self):

        if self.year <= 2006:
            case_dates = []
            for e in self.html.xpath(self.case_dates_path):
                month = ''.join(e.xpath(self.case_dates_path_month)).strip()
                day = ''.join(e.xpath(self.case_dates_path_day)).strip()
                month = month.replace('Decemberr', 'December')
                try:
                    case_dates.append(datetime.strptime('{} {} {}'.format(self.year, month, day), '%Y %B %d'))
                except ValueError:
                    raise
        else:
            case_dates = [parser.parse(''.join(i.strip() for i in d.xpath('.//text()')), fuzzy=True)
                          for d in self.html.xpath(self.case_dates_path)]

        return case_dates

    def _get_precedential_statuses(self):
        statuses = []
        for e in self.html.xpath(self.precedential_statuses_path):
            s = html.tostring(e, method='text', encoding='unicode')
            if 'NRel' in s:
                statuses.append('Unpublished')
            else:
                statuses.append('Published')
        return statuses

    def _get_docket_numbers(self):
        docket_numbers = []
        for e in self.html.xpath(self.docket_numbers_path):
            nums = ' '.join(i.strip() for i in e.xpath(".//text()"))
            nums = nums.replace('Official Reports', '')
            docket_numbers.append(nums)
        return docket_numbers

    def _get_neutral_citations(self):
        neutral_citations = None
        if self.neutral_citations_path:
            neutral_citations = []
            for e in self.html.xpath(self.neutral_citations_path):
                neutral_citations.append(' '.join(i.strip() for i in e.xpath(".//text()")).strip())

        return neutral_citations

    def _download_backwards(self, d):
        self.__set_paths(d)

        self.url = self.url_base.format(year=self.year)
        logger.info('Scraping year {}'.format(self.year))
        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200

    def __set_paths(self, year):
        self.year = year
        if year < 2006:
            base_path = '//tr/td[4][a]'
            self.download_url_path = '{}/a[1]/@href'.format(base_path)
            self.case_name_path = '{}/preceding-sibling::td[1]'.format(base_path)
            self.case_dates_path = base_path
            self.case_dates_path_day = './preceding-sibling::td[3]//text()'
            self.case_dates_path_month = './/ancestor::tr[1]/preceding-sibling::tr[count(./td) = 1][td//strong][1]/td//text()'
            self.precedential_statuses_path = '{}/preceding-sibling::td[2]'.format(base_path)
            self.docket_numbers_path = '{}/preceding-sibling::td[2]'.format(base_path)
            self.neutral_citations_path = None
        elif 2006 <= year <= 2010:
            if year == 2010:
                base_path = '//tr/td[3][div/a]'
                self.download_url_path = '{}/div[1]/a[1]/@href'.format(base_path)
                self.case_name_path = '{}/div[1]/a[1]'.format(base_path)
            else:
                base_path = '//tr/td[3][a]'
                self.download_url_path = '{}/a[1]/@href'.format(base_path)
                self.case_name_path = '{}/a[1]'.format(base_path)
            if year == 2006:
                self.case_dates_path = base_path
                self.case_dates_path_day = './preceding-sibling::td[2]//text()'
                self.case_dates_path_month = './/ancestor::tr[1]/preceding-sibling::tr[count(./td) = 1][1]/td//text()'
            else:
                self.case_dates_path = '{}/preceding-sibling::td[2]'.format(base_path)
            self.precedential_statuses_path = '{}/preceding-sibling::td[1]'.format(base_path)
            self.docket_numbers_path = '{}/preceding-sibling::td[1]'.format(base_path)
            self.neutral_citations_path = None
        elif 2010 < year < 2015:
            base_path = '//tr/td[4][div/a/text()]'
            self.download_url_path = '{}/div[1]/a[1]/@href'.format(base_path)
            self.case_name_path = '{}/div[1]/a[1]'.format(base_path)
            self.case_dates_path = '{}/preceding-sibling::td[3]'.format(base_path)
            self.precedential_statuses_path = '{}/preceding-sibling::td[2]'.format(base_path)
            self.docket_numbers_path = '{}/preceding-sibling::td[2]'.format(base_path)
            self.neutral_citations_path = '{}/preceding-sibling::td[1]'.format(base_path)



# History:
#   2015-10-20: Created by Andrei Chelaru

from juriscraper.opinions.united_states_backscrapers.state import ill


class Site(ill.Site):
    def __init__(self):
        super(Site, self).__init__()
        self.court_id = self.__module__
        self.url_base = 'http://www.illinoiscourts.gov/Opinions/AppellateCourt/{year}/4thDistrict/default.asp'



class SlownessException(Exception):
    """Raised when things are too slow."""

    def __init__(self, message):
        Exception.__init__(self, message)


class ParsingException(Exception):
    """Raised when parsing fails."""

    def __init__(self, message):
        Exception.__init__(self, message)


class InsanityException(Exception):
    """Raised when data validation fails."""

    def __init__(self, message):
        Exception.__init__(self, message)


class PacerLoginException(Exception):
    """Raised when the system cannot authenticate with PACER"""

    def __init__(self, message):
        Exception.__init__(self, message)



import re

import requests
import tldextract
from lxml import html
from six import string_types

from ..lib.string_utils import force_unicode


def get_pacer_court_info():
    r = requests.get("https://court-version-scraper.herokuapp.com/courts.json")
    return r.json()


def get_courts_from_json(j):
    courts = []
    for k, v in j.items():
        for court in v['courts']:
            court['type'] = k
            courts.append(court)
    return courts


def get_court_id_from_url(url):
    """Extract the court ID from the URL."""
    parts = tldextract.extract(url)
    return parts.subdomain.split('.')[1]


def get_pacer_case_id_from_nonce_url(url):
    """Extract the pacer case ID from the URL.

    In: https://ecf.almd.uscourts.gov/cgi-bin/DktRpt.pl?56120
    Out: 56120
    In: https://ecf.azb.uscourts.gov/cgi-bin/iquery.pl?625371913403797-L_9999_1-0-663150
    Out: 663150
    """
    param = url.split('?')[1]
    if 'L' in param:
        return param.rsplit('-', 1)[1]
    return param


def get_pacer_seq_no_from_doc1_url(url):
    """Extract the seq_no from the doc1 URL."""
    match = re.search('de_seq_num=(\d+)', url)
    if match:
        return match.group(1)
    else:
        return None


def get_pacer_doc_id_from_doc1_url(url):
    """Extract the pacer document ID from the doc1 URL. Coerce the fourth digit
    to zero.

    In:  https://ecf.almd.uscourts.gov/doc1/01712427473
    Out: 01702427473
    In:  /doc1/01712427473
    Out: 01702427473

    Note that results are strings, not ints, because many of the strings start
    with zero.

    See tests for more examples.
    """
    assert "show_case_doc" not in url, \
        "Cannot get doc1 ID from show_case_doc URL"
    url = url.rsplit('/', 1)[1].split('?')[0]
    url = url[:3] + "0" + url[4:]
    return url


def get_pacer_seq_no_from_doc1_anchor(anchor):
    """Extract the PACER sequence number from an HTML anchor node.

    :param anchor: An LXML Element.
    :return: None if no sequence number can be found. Otherwise returns the
    sequence number.
    """
    try:
        onclick = anchor.xpath('./@onclick')[0]
    except IndexError:
        return None
    else:
        if 'goDLS' in onclick:
            go_dls_parts = reverse_goDLS_function(onclick)
            return go_dls_parts['de_seq_num']


def reverse_goDLS_function(s):
    """Extract the arguments from the goDLS JavaScript function.

    In: goDLS('/doc1/01712427473','56121','69','','','1','','');return(false);
    Out: {
      'form_post_url': '/doc1/01712427473',
      'caseid': '56121',
      'de_seq_num': '69',
      'got_receipt': '',
      'pdf_header': '',
      'pdf_toggle_possible': '1',
      'magic_num': '',
      'hdr': '',
    }

    The key names correspond to the form field names in the JavaScript on PACER,
    but we don't actually know what each of these values does. Our best
    speculation is:

     - form_post_url: Where the form is posted to. The HTML 'action' attribute.
     - caseid: The internal PACER ID for the case.
     - de_seq_num: Unclear. This seems to be the internal ID for the document,
       but this field can be omitted without any known issues.
     - got_receipt: If set to '1', this will bypass the receipt page and
       immediately direct you to the page where the PDF is embedded in an
       iframe.
     - pdf_header: Can be either 1 or 2. 1: Show the header. 2: No header.
     - pdf_toggle_possible: This seems to always be 1. Could be that some courts
       do not allow the header to be turned off, but we haven't discoered that
       yet.
     - magic_num: This is used for the "One free look" downloads.
     - hdr: Unclear what HDR stands for but on items that have attachments,
       passing this parameter bypasses the download attachment screen and takes
       you directly to the PDF that you're trying to download. For an example,
       see document 108 from 1:12-cv-00102 in tnmd, which is a free opinion that
       has an attachment. Note that the eighth parameter was added some time
       after 2010. Dockets older than that date only have seven responses.
    """
    args = re.findall("\'(.*?)\'", s)
    parts = {
        'form_post_url': args[0],
        'caseid': args[1],
        'de_seq_num': args[2],
        'got_receipt': args[3],
        'pdf_header': args[4],
        'pdf_toggle_possible': args[5],
        'magic_num': args[6],
    }
    try:
        parts['hdr'] = args[7]
    except IndexError:
        # At some point dockets added this eighth parameter. Older ones lack it
        parts['hdr'] = None
    return parts


def make_doc1_url(court_id, pacer_doc_id, skip_attachment_page):
    """Make a doc1 URL.

    If skip_attachment_page is True, we replace the fourth digit with a 1
    instead of a zero, which bypasses the attachment page.
    """
    if skip_attachment_page and pacer_doc_id[3] == '0':
        # If the fourth digit is a 0, replace it with a 1
        pacer_doc_id = pacer_doc_id[:3] + '1' + pacer_doc_id[4:]
    return 'https://ecf.%s.uscourts.gov/doc1/%s' % (court_id,
                                                    pacer_doc_id)


def is_pdf(response):
    """Determines whether the item downloaded is a PDF or something else."""
    if response.headers.get('content-type') == 'application/pdf':
        return True
    return False


def get_nonce_from_form(r):
    """Get a nonce value from a HTML response. Returns the first nonce that is
    found.

    :param r: The response object you wish to parse.
    :returns A nonce object that can be used to query PACER or None, if no nonce
    can be found.
    """
    tree = html.fromstring(r.text)
    form_attrs = tree.xpath('//form//@action')
    for attr in form_attrs:
        # The action attr will be a value like:
        # ../cgi-bin/HistDocQry.pl?112801540788508-L_1_0-1
        # Split on the '?', and return the nonce.
        path, nonce = attr.split('?')
        if '-L_' in nonce:
            return nonce
    return None


def clean_pacer_object(obj):
    """Clean a list or dict that is part of a scraping response.

    PACER data is notoriously horrible, so this function attempts to clean up
    common problems that it may have. You can pass in either a dict or a list,
    and it will be cleaned recursively.

    Supported cleanup includes:

    1. Removing spaces before commas.
    1. Stripping whitespace from the ends.
    1. Normalizing white space.
    1. Forcing unicode.

    :param obj: A dict or list containing string objects.
    :return: A dict or list with the string values cleaned.
    """
    if isinstance(obj, list):
        l = []
        for i in obj:
            l.append(clean_pacer_object(i))
        return l
    elif isinstance(obj, dict):
        d = {}
        for k, v in obj.items():
            d[k] = clean_pacer_object(v)
        return d
    elif isinstance(obj, string_types):
        s = ' '.join(obj.strip().split())
        s = force_unicode(s)
        return re.sub('\s+,', ',', s)
    else:
        return obj


BASE_IA_URL = "https://www.archive.org/download"


def get_bucket_name(court, pacer_case_id):
    bucketlist = ["gov", "uscourts", court, unicode(pacer_case_id)]
    return ".".join(bucketlist)


def get_docket_filename(court, pacer_case_id):
    return ".".join(["gov", "uscourts", unicode(court), unicode(pacer_case_id),
                     "docket.xml"])


def get_document_filename(court, pacer_case_id, document_number,
                          attachment_number):
    return ".".join(["gov", "uscourts", unicode(court), unicode(pacer_case_id),
                     unicode(document_number), unicode(attachment_number or 0),
                     "pdf"])


def get_docketxml_url(court, pacer_case_id):
    return "%s/%s/%s" % (
        BASE_IA_URL,
        get_bucket_name(court, pacer_case_id),
        get_docket_filename(court, pacer_case_id),
    )


def get_pdf_url(court, pacer_case_id, document_number, attachment_number):
    return "%s/%s/%s" % (
        BASE_IA_URL,
        get_bucket_name(court, pacer_case_id),
        get_document_filename(court, pacer_case_id, document_number,
                              attachment_number),
    )



"""Scraper for the United States Department of Justice Attorney General
CourtID: ag
Court Short Name: United States Attorney General
"""

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.justice.gov/olc/opinions?items_per_page=40"
        self.back_scrape_url = self.url + '&page=%d'
        self.back_scrape_iterable = None
        self.cell_path = '//table//tr/td[%d]'

    def _download(self, request_dict={}):
        if not self.test_mode_enabled():
            # don't set this if running tests, as it hits the network
            self.back_scrape_iterable = range(0, self.get_last_page() + 1)
        return super(Site, self)._download(request_dict)

    def _get_case_names(self):
        cells = self.html.xpath(self.cell_path % 2)
        return [cell.text_content().strip() for cell in cells]

    def _get_download_urls(self):
        base_path = self.cell_path % 2
        return [href for href in self.html.xpath('%s//a[1]/@href' % base_path)]

    def _get_case_dates(self):
        cells = self.html.xpath(self.cell_path % 1)
        return [convert_date_string(cell.text_content().strip()) for cell in cells]

    def _get_docket_numbers(self):
        """advisory opinions have no docket numbers"""
        return [''] * len(self.case_names)

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_summaries(self):
        cells = self.html.xpath(self.cell_path % 3)
        return [cell.text_content().strip() for cell in cells]

    def get_last_page(self):
        html = self._get_html_tree_by_url(self.url)
        path = "//li[contains(@class, 'pager__item--last')]/a[1]/@href"
        return int(html.xpath(path)[0].split('=')[-1])

    def _download_backwards(self, page_num):
        self.url = self.back_scrape_url % page_num
        self.html = self._download()



'''
CourtID: illappct
Court Short Name: Ill. App. Ct.
Author: Rebecca Fordon
Reviewer: Mike Lissner
History:
* 2016-06-23: Created by Rebecca Fordon
'''

from juriscraper.oral_args.united_states.state import ill


class Site(ill.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.illinoiscourts.gov/Media/Appellate/5th_District.asp'
        self.download_url_path = '/td[5]//a[.//img]/@href'
        self.case_name_path = '/td[3]'
        self.docket_number_path = "/td[2]"
